# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum, auto
import dataclasses
from functools import wraps
import glob
import hashlib
import logging
import os.path
import time
from typing import (
    Optional,
    AbstractSet,
    MutableSet,
    Any,
    Callable,
    NamedTuple,
    TYPE_CHECKING,
)

from pisek.jobs.cache import Cache, CacheEntry
from pisek.utils.paths import TaskPath

if TYPE_CHECKING:
    from pisek.env.env import Env

logger = logging.getLogger(__name__)


class State(Enum):
    in_queue = auto()
    running = auto()
    succeeded = auto()
    failed = auto()
    cancelled = auto()

    def finished(self) -> bool:
        return self in (State.succeeded, State.failed, State.cancelled)


class PipelineItemFailure(Exception):
    pass


class PipelineItemAbort(Exception):
    def __init__(self, item: "PipelineItem"):
        super().__init__(f'"{item.name}" was aborted.')


class CaptureInitParams:
    """
    Class that stores __init__ args and kwargs of its descendants
    and gets accessed Env reads. Only does that to the topmost __init__.
    """

    _initialized = False
    _accessed_envs: MutableSet[tuple[str, ...]]

    def __init_subclass__(cls):
        real_init = cls.__init__

        @wraps(real_init)
        def wrapped_init(self, env: "Env", *args, **kwargs):
            toplevel = not self._initialized
            if toplevel:
                self._args = args
                self._kwargs = kwargs
                self._initialized = True
                self._env = env
                self._env.clear_accesses()

            real_init(self, self._env, *args, **kwargs)

            if toplevel:
                self._accessed_envs |= self._env.get_accessed()
                self._env.clear_accesses()

        cls.__init__ = wrapped_init


class RequiredBy(NamedTuple):
    pipeline_item: "PipelineItem"
    name: Optional[str]
    run_condition: Callable[[Any], bool]


class PipelineItem(ABC):
    """Generic PipelineItem with state and dependencies."""

    _env: "Env"
    run_always: bool = False  # Runs even if prerequisites failed

    def __init__(self, name: str) -> None:
        self.name = name
        self.state = State.in_queue
        self.result: Optional[Any] = None
        self.fail_msg = ""

        self.prerequisites = 0
        self.required_by: list[RequiredBy] = []
        self.prerequisites_results: dict[str, Any] = {}
        # List of prints (string to print, whether to use stderr)
        self.terminal_output: list[tuple[str, bool]] = []

    def _colored(self, msg: str, color: str) -> str:
        return self._env.colored(msg, color)

    def _print(self, msg: str, end: str = "\n", stderr: bool = False) -> None:
        """Adds text for printing to stdout/stderr later."""
        self.terminal_output.append((msg + end, stderr))

    def _warn(self, msg: str) -> None:
        if self._env.strict:
            raise PipelineItemFailure(msg)
        else:
            self._print(self._colored(msg, "yellow"))

    def _fail(self, failure: PipelineItemFailure) -> None:
        """End this job in failure."""
        if self.fail_msg != "":
            raise RuntimeError("PipelineItem cannot fail twice.")
        self.state = State.failed
        self.fail_msg = str(failure)

    def cancel(self) -> None:
        """Cancels job and all that require it."""
        if self.state in (State.succeeded, State.cancelled):
            return  # No need to cancel
        self.state = State.cancelled
        for item, _, _ in self.required_by:
            if not item.run_always:
                item.cancel()

    def _check_prerequisites(self) -> None:
        """Checks if all prerequisites are finished raises error otherwise."""
        if not self.run_always and self.prerequisites > 0:
            raise RuntimeError(
                f"{self.__class__.__name__} {self.name} prerequisites not finished ({self.prerequisites} remaining)."
            )

    def add_prerequisite(
        self,
        item: Optional["PipelineItem"],
        name: Optional[str] = None,
        condition: Callable[[Any], bool] = lambda _: True,
    ) -> None:
        """Adds given PipelineItem as a prerequisite to this job."""
        if item is None:
            return

        self.prerequisites += 1
        item.required_by.append(RequiredBy(self, name, condition))

    def finish(self) -> None:
        """Notifies PipelineItems that depend on this job."""
        for item, name, condition in self.required_by:
            if item.run_always or (
                self.state == State.succeeded and condition(self.result)
            ):
                item.prerequisites -= 1
                if name is not None:
                    item.prerequisites_results[name] = deepcopy(self.result)
            else:
                item.cancel()


class Job(PipelineItem, CaptureInitParams):
    """One simple cacheable task in pipeline."""

    _args: list[Any]
    _kwargs: dict[str, Any]

    def __init__(self, env: "Env", name: str) -> None:
        self._env = env
        self._accessed_envs: MutableSet[tuple[str, ...]] = set()
        self._accessed_globs: MutableSet[str] = set()
        self._accessed_files: MutableSet[str] = set()
        self._logs: list[tuple[str, str]] = []
        self.name = name
        self.started: float | None = None
        super().__init__(name)

    def _log(self, kind: str, message: str) -> None:
        self._logs.append((kind, message))
        getattr(logger, kind)(message)

    def _access_file(self, filename: TaskPath) -> None:
        """Add file this job depends on."""
        self._accessed_files.add(filename.path)

    @property
    def accessed_files(self) -> set[str]:
        return set(self._accessed_files)

    def _signature(
        self,
        envs: AbstractSet[tuple[str, ...]],
        paths: AbstractSet[str],
        globs: AbstractSet[str],
        results: dict[str, Any],
        cache: Cache,
    ) -> tuple[Optional[str], Optional[str]]:
        """Compute a signature (i.e. hash) of given envs, files and prerequisites results."""
        sign = hashlib.sha256()
        sign.update(f"{self.__class__.__name__}\n".encode())
        for i, arg in enumerate(self._args):
            sign.update(f"{i}={arg}\n".encode())
        for key, val in self._kwargs.items():
            sign.update(f"{key}={val}\n".encode())

        for env_key in sorted(envs):
            try:
                value = self._env.get_compound(env_key)
            except (AttributeError, TypeError, ValueError, KeyError):
                return (None, f"Key nonexistent: {env_key}")
            sign.update(f"{env_key}={value}\n".encode())

        for path in sorted(paths):
            while os.path.islink(path):
                path = os.path.normpath(
                    os.path.join(os.path.dirname(path), os.readlink(path))
                )

            if os.path.isfile(path):
                sign.update(f"{path}={cache.file_hash(path)}\n".encode())
            elif os.path.isdir(path):
                sign.update(f"{path} is directory\n".encode())
            else:
                return (None, f"File nonexistent: {path}")

        for g in sorted(globs):
            glob_sign = f"{g} -> " + " ".join(
                glob.glob(g, recursive=True, include_hidden=True)
            )
            sign.update(glob_sign.encode())

        for name, result in sorted(results.items()):
            # Trying to prevent hashing object.__str__ which is non-deterministic
            assert (
                result is None
                or isinstance(result, (str, int, float))
                or dataclasses.is_dataclass(result)
            )
            sign.update(f"{name}={result}\00".encode())

        return (sign.hexdigest(), None)

    def _find_entry(self, cache: Cache) -> Optional[CacheEntry]:
        """Finds a corresponding CacheEntry for this Job."""
        for cache_entry in cache[self.name]:
            sign, err = self._signature(
                set(cache_entry.envs),
                set(cache_entry.files),
                set(cache_entry.globs),
                self.prerequisites_results,
                cache,
            )
            if cache_entry.signature == sign:
                return cache_entry
        return None

    def _export(self, result: Any, cache: Cache) -> CacheEntry:
        """Export this job into CacheEntry."""
        sign, err = self._signature(
            self._accessed_envs,
            self._accessed_files,
            self._accessed_globs,
            self.prerequisites_results,
            cache,
        )
        if sign is None:
            raise RuntimeError(
                f"Computing signature of job {self.name} failed:\n  {err}."
            )
        return CacheEntry(
            self.name,
            sign,
            result,
            self._accessed_envs,
            self._accessed_files,
            self._accessed_globs,
            self.prerequisites_results,
            self.terminal_output,
            self._logs,
        )

    def prepare(self, cache: Cache | None) -> None:
        if self.state == State.cancelled:
            return None
        self._check_prerequisites()

        if (
            cache is not None
            and self.name in cache
            and (entry := self._find_entry(cache))
        ):
            logger.info(f"Loading cached '{self.name}'")
            cache.move_to_top(entry)
            self.terminal_output = entry.output
            for level, message in entry.logs:
                getattr(logger, level)(message)
            self._accessed_files = set(entry.files)
            self.result = entry.result
            self.state = State.succeeded

    def run(self, env: "Env") -> None:
        """Run this job."""
        if self.state == State.cancelled:
            return
        self.state = State.running
        self.started = time.time()
        logger.info(f"Running '{self.name}'")

        try:
            self._env = env
            self._env.clear_accesses()
            self.result = self._run()
            self._accessed_envs |= self._env.get_accessed()
        except PipelineItemFailure as failure:
            self._fail(failure)

    def finalize(self, cache: Cache | None):
        if self.state == State.running:
            if cache is not None:
                cache.add(self._export(self.result, cache))
            self.state = State.succeeded
        return self.finish()

    @abstractmethod
    def _run(self) -> Any:
        """What this job actually does (without all the management)."""
        pass


class JobManager(PipelineItem):
    """Object that can create jobs and compute depending on their results."""

    def set_env(self, env: "Env") -> None:
        self._env = env

    def create_jobs(self) -> list[Job]:
        """Crates this JobManager's jobs."""
        self.result: Optional[dict[str, Any]]
        if self.state == State.cancelled:
            self.jobs = []
        else:
            self.state = State.running
            self._check_prerequisites()
            try:
                self.jobs = self._get_jobs()
            except PipelineItemFailure as failure:
                self._fail(failure)
                self.jobs = []

        return self.jobs

    @abstractmethod
    def _get_jobs(self) -> list[Job]:
        """Actually creates this JobManager's jobs (without management)."""
        pass

    def _job_states(self) -> tuple[State, ...]:
        """States of this manager's jobs."""
        return tuple(map(lambda j: j.state, self.jobs))

    def jobs_with_state(self, state: State) -> list[Job]:
        """Filter this manager's jobs by state."""
        return list(filter(lambda j: j.state == state, self.jobs))

    def update(self) -> None:
        """Update this manager's state according to its jobs."""
        pass

    @abstractmethod
    def get_status(self) -> str:
        """Return status of job manager to be displayed on stdout."""
        return ""

    def ready(self) -> bool:
        """
        Returns whether manager is ready for evaluation.
        (i.e. All of it's jobs have finished)
        """
        return self.state == State.running and len(
            self.jobs_with_state(State.succeeded)
            + self.jobs_with_state(State.cancelled)
        ) == len(self.jobs)

    def any_failed(self) -> bool:
        """Returns whether this manager or its jobs had any failures so far."""
        return self.state == State.failed or len(self.jobs_with_state(State.failed)) > 0

    def finalize(self) -> None:
        """Does final evaluation and computes the result of this job manager."""
        if not self.any_failed():
            try:
                self._evaluate()
            except PipelineItemFailure as failure:
                self._fail(failure)
            else:
                self.state = State.succeeded
        else:
            self.state = State.failed

        self.result = self._compute_result()
        super().finish()

    def _evaluate(self) -> None:
        """Decide whether jobs did run as expected and return result."""
        pass

    def _compute_result(self) -> dict[str, Any]:
        """Creates result to be read by other managers."""
        return {}



#!/usr/bin/env python3
# For testing that a Python generator works as well

import random
import sys

diff = int(sys.argv[1]) - 1
seed = int(sys.argv[2], 16)
random.seed(seed)

MAX_ABS = [int(1e9), int(1e18)][diff]

T = 10
print(T)

for ti in range(T):
    a = random.randint(-MAX_ABS, MAX_ABS)
    b = random.randint(-MAX_ABS, MAX_ABS)
    print(a, b)



#!/usr/bin/env python3
import os
import sys


def fail(msg):
    print(msg, file=sys.stderr)
    exit(43)


def main():
    # Check that the command line arguments are passed as expected,
    # even though we do not need them here otherwise.
    assert len(sys.argv) == 3, "Expected two arguments: $difficulty $seed"

    diff = int(sys.argv[1])
    assert diff in [0, 1, 2]

    # Check seed conforms to specification
    seed = sys.argv[2]
    if seed == "-":
        assert diff == 0
    else:
        _seed = int(seed, 16)

    test_input_file = os.getenv("TEST_INPUT")
    test_output_file = os.getenv("TEST_OUTPUT")

    if test_input_file is None or test_output_file is None:
        exit(2)

    with open(test_input_file, "r") as fin:
        with open(test_output_file, "r") as fcorrect:
            t = int(fin.readline())

            for _ in range(t):
                a, b = [int(x) for x in fin.readline().split()]
                c = int(fcorrect.readline())

                # Normally we would not need to do this in the judge, but here we want
                # to verify that the command line arguments are passed as expected.
                if diff == 1:
                    assert abs(a) <= 1e9, "Input out of bounds for the easy version"
                    assert abs(b) <= 1e9, "Input out of bounds for the easy version"

                try:
                    contestant = int(input())
                except ValueError:
                    fail("Not a number")
                except EOFError:
                    fail("Output ends unexpectedly")

                assert a + b == c

                if c != contestant:
                    fail("Wrong answer")

    try:
        input()
        fail("Trailing output")
    except EOFError:
        pass

    print("Correct answer", file=sys.stderr)
    exit(42)


main()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from decimal import Decimal
from typing import Any, Optional

from pisek.utils.text import pad, tab
from pisek.utils.terminal import right_aligned_text
from pisek.utils.colors import ColorSettings
from pisek.utils.paths import IInputPath

from pisek.jobs.jobs import State, Job, PipelineItemFailure
from pisek.env.env import Env
from pisek.config.config_types import TaskType
from pisek.task_jobs.tools import SanitizeAbstract
from pisek.task_jobs.data.data import SymlinkData
from pisek.task_jobs.solution.verdicts_eval import check_verdicts, compute_verdict
from pisek.task_jobs.task_job import TaskHelper
from pisek.task_jobs.task_manager import TaskJobManager
from pisek.task_jobs.data.testcase_info import TestcaseInfo, TestcaseGenerationMode
from pisek.task_jobs.generator.generator_manager import TestcaseInfoMixin
from pisek.task_jobs.solution.solution_result import Verdict, SolutionResult
from pisek.task_jobs.solution.solution import (
    RunSolution,
    RunBatchSolution,
    RunInteractive,
)
from pisek.task_jobs.checker.checker import checker_job
from pisek.task_jobs.checker.cms_judge import RunCMSJudge
from pisek.task_jobs.checker.checker_base import RunChecker, RunBatchChecker


class SolutionManager(TaskJobManager, TestcaseInfoMixin):
    """Runs a solution and checks if it works as expected."""

    def __init__(self, solution_label: str, generate_inputs: bool) -> None:
        self.solution_label: str = solution_label
        self._generate_inputs = generate_inputs
        self.solution_points: Optional[Decimal] = None
        self.tests: list[TestJobGroup] = []
        self._tests_results: dict[int, Verdict] = {}
        super().__init__(f"Test {solution_label}")

    def _get_jobs(self) -> list[Job]:
        self._solution = self._env.config.solutions[self.solution_label]
        self.is_primary: bool = self._solution.primary

        self._sols: dict[IInputPath, RunSolution] = {}
        self._checkers: dict[IInputPath, RunChecker] = {}
        self._static_out_checkers: dict[IInputPath, RunChecker] = {}

        for sub_num, inputs in self._all_testcases().items():
            self.tests.append(
                TestJobGroup(self._env, sub_num, self._solution.tests[sub_num])
            )
            for inp in inputs:
                self._add_testcase_info_jobs(inp, sub_num)

        return self._jobs

    def _register_skipped_testcase(
        self, testcase_info: TestcaseInfo, seed: Optional[int], test: int
    ) -> None:
        super()._register_skipped_testcase(testcase_info, seed, test)
        input_path = testcase_info.input_path(seed, solution=self.solution_label)
        if self._env.config.test_sections[test].new_in_test(input_path.name):
            self.tests[-1].new_run_jobs.append(self._sols[input_path])
            self.tests[-1].new_jobs.append(self._checkers[input_path])
            self._sols[input_path].require()
        else:
            self.tests[-1].previous_jobs.append(self._checkers[input_path])

    def _add_generate_input_jobs(
        self,
        testcase_info: TestcaseInfo,
        seed: Optional[int],
        test: int,
        test_determinism: bool,
    ) -> None:
        if self._generate_inputs:
            super()._add_generate_input_jobs(
                testcase_info, seed, test, test_determinism
            )

        self._add_job(
            SymlinkData(
                self._env,
                testcase_info.input_path(seed),
                testcase_info.input_path(seed, solution=self.solution_label),
            ),
            new_last=True,
        )

    def _add_respects_seed_jobs(
        self, testcase_info: TestcaseInfo, seeds: list[int], test: int
    ) -> None:
        if self._generate_inputs:
            super()._add_respects_seed_jobs(testcase_info, seeds, test)

    def _add_solution_jobs(
        self,
        testcase_info: TestcaseInfo,
        seed: Optional[int],
        test: int,
    ) -> None:
        input_path = testcase_info.input_path(seed, solution=self.solution_label)

        run_sol: RunSolution
        run_checker: RunChecker
        if self._env.config.task.task_type == TaskType.batch:
            if (
                testcase_info.generation_mode == TestcaseGenerationMode.static
                and self._generate_inputs
            ):
                # Check static outputs against themselves
                inp = testcase_info.input_path(seed)
                out = testcase_info.reference_output(self._env, seed)

                checker_j = checker_job(
                    inp, out, out, test, seed, Verdict.ok, self._env
                )
                self._add_check_output_jobs(out)
                self._add_job(
                    checker_j,
                    prerequisite_name=(
                        "sanitize"
                        if isinstance(self._testcase_last, SanitizeAbstract)
                        else None
                    ),
                )
                self._static_out_checkers[inp] = checker_j

            run_batch_sol, run_checker = self._create_batch_jobs(
                testcase_info, seed, test
            )
            run_sol = run_batch_sol
            self._add_job(run_batch_sol, new_last=True)

            self._add_check_output_jobs(run_batch_sol.output.to_sanitized_output())

            if isinstance(self._testcase_last, SanitizeAbstract):
                run_checker.add_prerequisite(self._testcase_last, "sanitize")

            if self._env.config.tests.judge_needs_out:
                self._add_job(
                    SymlinkData(
                        self._env,
                        testcase_info.reference_output(self._env, seed),
                        testcase_info.reference_output(
                            self._env, seed, solution=self.solution_label
                        ),
                    ),
                    new_last=True,
                )

            self._add_job(run_checker)

        elif self._env.config.task.task_type == TaskType.interactive:
            run_sol = run_checker = self._create_interactive_jobs(input_path, test)
            self._add_job(run_sol)

        self._sols[input_path] = run_sol
        self._checkers[input_path] = run_checker
        self.tests[-1].new_jobs.append(run_checker)
        self.tests[-1].new_run_jobs.append(run_sol)

    def _create_batch_jobs(
        self, testcase_info: TestcaseInfo, seed: Optional[int], test: int
    ) -> tuple[RunBatchSolution, RunBatchChecker]:
        """Create RunSolution and RunBatchChecker jobs for batch task type."""
        input_path = testcase_info.input_path(seed, solution=self.solution_label)
        run_solution = RunBatchSolution(
            self._env,
            self._solution.run,
            self.is_primary,
            input_path,
            input_path.to_output(),
        )

        out = input_path.to_output()
        run_checker = checker_job(
            input_path,
            out,
            testcase_info.reference_output(
                self._env,
                seed,
                solution=(
                    self.solution_label
                    if self._env.config.tests.judge_needs_out
                    else None
                ),
            ),
            test,
            seed,
            None,
            self._env,
        )
        run_checker.add_prerequisite(run_solution, name="run_solution")

        return (run_solution, run_checker)

    def _create_interactive_jobs(self, inp: IInputPath, test: int) -> RunInteractive:
        """Create RunInteractive job for interactive task type."""
        if self._env.config.tests.out_judge is None:
            raise RuntimeError("Unset judge for interactive.")

        return RunInteractive(
            self._env,
            self._solution.run,
            self.is_primary,
            self._env.config.tests.out_judge,
            test,
            inp,
        )

    def update(self):
        """Cancel running on inputs that can't change anything."""
        for test in self.tests:
            if test.definitive():
                test.cancel()

    def get_status(self) -> str:
        msg = self.solution_label
        if self.state == State.cancelled:
            return self._job_bar(msg)

        points = self._format_points(self.solution_points)
        total_points = self._format_points(self._env.config.total_points)

        max_time_f = max((s.slowest_time for s in self.tests), default=0.0)
        max_time = self._format_time(max_time_f)

        if not self.state.finished() or self._env.verbosity == 0:
            header = self._solution_header_verbosity0(
                msg, self.solution_points, max_time_f
            )
            tests_text = "".join(sub.status_verbosity0() for sub in self.tests)
        else:
            header = (
                right_aligned_text(
                    f"{msg}: {points}/{total_points}", f"slowest {max_time}"
                )
                + "\n"
            )
            header = self._colored(header, "cyan")
            tests_text = tab(
                "\n".join(
                    sub.status(self.tests, self._env.verbosity) for sub in self.tests
                )
            )
            if self._env.verbosity == 1:
                tests_text += "\n"

        return header + tests_text

    def _evaluate(self) -> None:
        """Evaluates whether solution preformed as expected."""
        self.solution_points = Decimal(0)
        for sub_job in self.tests:
            self.solution_points += sub_job.points
            self._tests_results[sub_job.num] = sub_job.verdict

        solution_conf = self._env.config.solutions[self.solution_label]
        for sub_job in self.tests:
            sub_job.assert_as_expected()

        points = solution_conf.points
        p_min = solution_conf.points_min
        p_max = solution_conf.points_max

        if points is not None and self.solution_points != points:
            raise PipelineItemFailure(
                f"Solution {self.solution_label} should have gotten {points} but got {self.solution_points} points."
            )
        elif p_min is not None and self.solution_points < p_min:
            raise PipelineItemFailure(
                f"Solution {self.solution_label} should have gotten at least {p_min} but got {self.solution_points} points."
            )
        elif p_max is not None and self.solution_points > p_max:
            raise PipelineItemFailure(
                f"Solution {self.solution_label} should have gotten at most {p_max} but got {self.solution_points} points."
            )

    def _compute_result(self) -> dict[str, Any]:
        result: dict[str, Any] = super()._compute_result()

        def add_checker_out(cj: RunChecker) -> None:
            if cj.result is None or cj.result.verdict not in (
                Verdict.ok,
                Verdict.partial_ok,
                Verdict.wrong_answer,
            ):
                return

            if isinstance(checker_job, RunCMSJudge):
                result["checker_outs"].add(cj.points_file)
            result["checker_outs"].add(cj.checker_log_file)

        result["results"] = {}
        result["checker_outs"] = set()
        for inp, checker_job in self._checkers.items():
            result["results"][inp] = checker_job.result
            add_checker_out(checker_job)

        for checker_job in self._static_out_checkers.values():
            add_checker_out(checker_job)

        result["tests"] = self._tests_results

        return result


class TestJobGroup(TaskHelper):
    """Groups jobs of a single test."""

    def __init__(self, env: Env, num: int, expected_str: str) -> None:
        self._env = env
        self.num = num
        self.test = env.config.test_sections[num]
        self.expected_str = expected_str

        self.new_run_jobs: list[RunSolution] = []
        self.previous_jobs: list[RunChecker] = []
        self.new_jobs: list[RunChecker] = []

        self._canceled: bool = False

    @property
    def all_jobs(self) -> list[RunChecker]:
        return self.previous_jobs + self.new_jobs

    @property
    def points(self) -> Decimal:
        results = self._results(self.all_jobs)
        points = map(lambda r: r.points(self._env, self.test.points), results)
        return min(points, default=Decimal(self.test.max_points))

    @property
    def verdict(self) -> Verdict:
        return compute_verdict(self._verdicts(self.all_jobs))

    @property
    def slowest_time(self) -> float:
        results = self._results(self.all_jobs)
        times = map(lambda r: r.solution_rr.time, results)
        return max(times, default=0.0)

    def _job_results(self, jobs: list[RunChecker]) -> list[Optional[SolutionResult]]:
        return list(map(lambda j: j.result, jobs))

    def _finished_jobs(self, jobs: list[RunChecker]) -> list[RunChecker]:
        return list(filter(lambda j: j.result is not None, jobs))

    def _results(self, jobs: list[RunChecker]) -> list[SolutionResult]:
        filtered = []
        for res in self._job_results(jobs):
            if res is not None:
                filtered.append(res)
        return filtered

    def _verdicts(self, jobs: list[RunChecker]) -> list[Verdict]:
        return list(map(lambda r: r.verdict, self._results(jobs)))

    def status(
        self, all_tests: list["TestJobGroup"], verbosity: Optional[int] = None
    ) -> str:
        verbosity = self._env.verbosity if verbosity is None else verbosity

        if verbosity <= 0:
            return self.status_verbosity0()
        elif verbosity == 1:
            return self.status_verbosity1()
        elif verbosity >= 2:
            return self.status_verbosity2(all_tests)

        raise RuntimeError(f"Unknown verbosity {verbosity}")

    def _verdict_marks(self, jobs: list[RunChecker]) -> str:
        return "".join(job.verdict_mark() for job in jobs)

    def _predecessor_summary(self) -> str:
        if not self.previous_jobs:
            return ""

        verdicts = self._verdicts(self.previous_jobs)
        if not verdicts:
            return "p |"

        _, _, guarantor = self._as_expected(self.previous_jobs)
        if guarantor is not None:
            assert guarantor.result is not None
            verdict = guarantor.result.verdict
        else:
            verdict = compute_verdict(verdicts)

        return f"p{verdict.mark}|"

    def status_verbosity0(self) -> str:
        left_bracket = "["
        right_bracket = "]"
        if self.definitive():
            color = self.verdict.color
            left_bracket = ColorSettings.colored(left_bracket, color)
            right_bracket = ColorSettings.colored(right_bracket, color)

        return (
            left_bracket
            + self._predecessor_summary()
            + self._verdict_marks(self.new_jobs)
            + right_bracket
        )

    def status_verbosity1(self) -> str:
        max_sub_name_len = max(
            len(test.name) for test in self._env.config.test_sections.values()
        )
        max_sub_points_len = max(
            len(self._format_points(sub.max_points))
            for sub in self._env.config.test_sections.values()
        )

        return right_aligned_text(
            f"{self.test.name:<{max_sub_name_len}}  "
            f"{self._format_points(self.points):>{max_sub_points_len}}  "
            f"{self.status_verbosity0()}",
            f"slowest {self._format_time(self.slowest_time)}",
            offset=-2,
        )

    def status_verbosity2(self, all_tests: list["TestJobGroup"]):
        def test_name(num: int) -> str:
            return self._env.config.test_sections[num].name

        text = ""
        max_inp_name_len = max(
            (len(j.input.name) for j in self.new_jobs if j.result is not None),
            default=0,
        )
        test_info = (
            right_aligned_text(
                f"{self.test.name}: {self._format_points(self.points)}/{self._format_points(self.test.max_points)}",
                f"slowest {self._format_time(self.slowest_time)}",
                offset=-2,
            )
            + "\n"
        )
        text += self._env.colored(test_info, "magenta")

        max_pred_name_len = max(
            (len(test_name(pred)) for pred in self.test.all_predecessors),
            default=0,
        )
        for pred in self.test.all_predecessors:
            pred_group = all_tests[pred]
            text += right_aligned_text(
                tab(
                    f"Predecessor {pad(test_name(pred) + ':', max_pred_name_len + 1)}  "
                    f"{pred_group.status_verbosity0()}"
                ),
                f"slowest {self._format_time(pred_group.slowest_time)}",
                offset=-2,
            )
            text += "\n"

        if len(self.test.all_predecessors) and any(
            map(lambda j: j.result, self.new_jobs)
        ):
            text += "\n"

        for job in self.new_jobs:
            if job.result is not None:
                points = ColorSettings.colored(
                    f"({self._format_points(job.result.points(self._env, self.test.points))})",
                    job.result.verdict.color,
                )
                input_verdict = tab(
                    f"{job.input.name:<{max_inp_name_len}} {points}: {job.verdict_text()}"
                )
                text += right_aligned_text(
                    input_verdict,
                    self._format_time(job.result.solution_rr.time),
                    offset=-2,
                )
                text += "\n"

        return text

    def definitive(self) -> bool:
        """Checks whether test jobs have resulted in outcome that cannot be changed."""
        if len(self._results(self.all_jobs)) == len(self.all_jobs):
            return True

        if self._env.all_inputs:
            return False

        if self.expected_str == "X" and not self.verdict.is_zero_point():
            return False  # Cause X is very very special

        return self._as_expected(self.all_jobs)[1]

    def assert_as_expected(self) -> None:
        """Checks this test resulted as expected. Raises PipelineItemFailure otherwise."""
        ok, _, breaker = self._as_expected(self.all_jobs)
        if not ok:
            msg = f"{self.test.name} did not result as expected: '{self.expected_str}'"
            if breaker is not None:
                msg += f"\n{tab(breaker.message())}"
            raise PipelineItemFailure(msg)

    def _as_expected(
        self, jobs: list[RunChecker]
    ) -> tuple[bool, bool, Optional[RunChecker]]:
        finished_jobs = self._finished_jobs(jobs)
        verdicts = self._results(jobs)

        result, definitive, guarantor = check_verdicts(
            list(map(lambda r: r.verdict, verdicts)), self.expected_str
        )

        guarantor_job = None if guarantor is None else finished_jobs[guarantor]

        return result, definitive, guarantor_job

    def cancel(self):
        if self._canceled:
            return

        self._canceled = True
        for job in self.new_run_jobs:
            job.unrequire()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from typing import Optional

from pisek.env.env import Env
from pisek.jobs.jobs import PipelineItemFailure
from pisek.config.task_config import RunSection
from pisek.utils.paths import IInputPath
from pisek.task_jobs.task_job import TaskJob
from pisek.task_jobs.program import ProgramsJob
from pisek.task_jobs.data.testcase_info import TestcaseInfo, TestcaseGenerationMode


class GeneratorListInputs(ProgramsJob):
    """Lists all inputs generator can generate."""

    def __init__(
        self, env: Env, generator: RunSection, *, name: str = "", **kwargs
    ) -> None:
        self.generator = generator
        super().__init__(env=env, name=name or "List generator inputs", **kwargs)

    @abstractmethod
    def _run(self) -> list[TestcaseInfo]:
        pass


class GenerateInput(ProgramsJob):
    """Generates input with given name."""

    def __init__(
        self,
        env: Env,
        generator: RunSection,
        testcase_info: TestcaseInfo,
        seed: Optional[int],
        input_path: IInputPath,
        *,
        name: str = "",
        **kwargs,
    ) -> None:
        assert testcase_info.generation_mode == TestcaseGenerationMode.generated

        self.generator = generator
        self.seed = seed
        self.testcase_info = testcase_info
        self.input_path = input_path
        super().__init__(
            env=env, name=name or f"Generate {self.input_path.name}", **kwargs
        )

    def _run(self) -> None:
        self._gen()

    @abstractmethod
    def _gen(self) -> None:
        pass


class GeneratorTestDeterminism(ProgramsJob):
    """Tests determinism of generating a given input."""

    def __init__(
        self,
        env: Env,
        generator: RunSection,
        testcase_info: TestcaseInfo,
        seed: Optional[int],
        *,
        name: str = "",
        **kwargs,
    ) -> None:
        assert testcase_info.generation_mode == TestcaseGenerationMode.generated

        self.generator = generator
        self.seed = seed
        self.testcase_info = testcase_info
        self._original = testcase_info.input_path(seed)
        self.input_path = self._original.to_second()
        super().__init__(
            env=env,
            name=name or f"Generator is deterministic (on {self._original:p})",
            **kwargs,
        )

    def _run(self) -> None:
        input_path = self.input_path.to_raw(self._env.config.tests.in_format)
        original = self._original.to_raw(self._env.config.tests.in_format)
        self._gen()
        if not self._files_equal(input_path, original):
            raise PipelineItemFailure(
                f"Generator is not deterministic. Files {input_path:p} and {original:p} differ"
                + (f" (seed {self.seed:016x})" if self.testcase_info.seeded else "")
                + "."
            )
        self._remove_file(input_path)

    @abstractmethod
    def _gen(self) -> None:
        pass


class GeneratorRespectsSeed(TaskJob):
    def __init__(
        self, env: Env, testcase_info: TestcaseInfo, seed1: int, seed2: int
    ) -> None:
        assert (
            testcase_info.generation_mode == TestcaseGenerationMode.generated
            and testcase_info.seeded
        )

        self.testcase_info = testcase_info
        self.seed1 = seed1
        self.seed2 = seed2
        self.input1 = testcase_info.input_path(seed1).to_raw(env.config.tests.in_format)
        self.input2 = testcase_info.input_path(seed2).to_raw(env.config.tests.in_format)
        super().__init__(
            env=env,
            name=f"Generator respects seed ({self.input1:n} and {self.input2:n})",
        )

    def _run(self) -> None:
        if self._files_equal(self.input1, self.input2):
            raise PipelineItemFailure(
                f"Inputs generated with different seed are same:\n"
                f"- {self.input1:p} (seed {self.seed1:016x})\n"
                f"- {self.input2:p} (seed {self.seed2:016x})"
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from configparser import (
    ConfigParser,
    DuplicateSectionError,
    DuplicateOptionError,
    MissingSectionHeaderError,
)
from dataclasses import dataclass
from importlib.resources import files
import os
import re
from typing import Optional, Iterable

from pisek.utils.text import tab
from pisek.user_errors import TaskConfigError, TaskConfigParsingError

from .update_config import update_config
from .config_description import ConfigKeysHelper

GLOBAL_DEFAULTS = "global-defaults"
GLOBAL_DEFAULTS_FILE = str(files("pisek").joinpath("config", GLOBAL_DEFAULTS))
V2_DEFAULTS = {
    task_type: str(files("pisek").joinpath(f"config/{task_type}-defaults"))
    for task_type in ["kasiopea", "cms"]
}

DEFAULT_CONFIG_FILENAME = "config"


def new_config_parser() -> ConfigParser:
    config = ConfigParser(interpolation=None)
    config.optionxform = lambda x: x  # type: ignore
    return config


@dataclass(frozen=True)
class ConfigValue:
    value: str
    config: str
    section: str
    key: Optional[str]
    internal: bool = False

    @staticmethod
    def make_internal(value: str, section: str, key: Optional[str] = None):
        return ConfigValue(value, "_internal", section, key, True)

    def location(self) -> str:
        text = f"section [{self.section}]"
        if self.key is not None:
            text += f", key '{self.key}'"

        if self.internal:
            text += " (internal pisek value)"
        else:
            text += f" (in config file {os.path.basename(self.config)})"

        return text

    def split(self, sep: Optional[str] = None) -> list["ConfigValue"]:
        return [
            ConfigValue(part, self.config, self.section, self.key, self.internal)
            for part in self.value.split(sep=sep)
        ]

    def change_value(self, new_value: str) -> "ConfigValue":
        return ConfigValue(
            new_value, self.config, self.section, self.key, self.internal
        )


class ConfigHierarchy:
    """Represents hierarchy of config files where last overrides all previous."""

    def __init__(
        self,
        task_path: str,
        info: bool,
        pisek_directory: Optional[str],
        config_filename: str,
    ) -> None:
        self._task_path = task_path
        self._pisek_directory = pisek_directory

        self._config_paths: list[str] = []
        self._configs: list[ConfigParser] = []

        # We want set that stores the order of insertions
        self.loaded_values: dict[ConfigValue, None] = {}

        self._load_config(os.path.join(task_path, config_filename), info)
        self._load_config(GLOBAL_DEFAULTS_FILE, False)

    def _load_config(self, path: str, info: bool = True) -> None:
        self._config_paths.append(path)
        self._configs.append(config := new_config_parser())
        if not self._read_config(config, path):
            raise TaskConfigError(f"Missing config {path}. Is this task folder?")

        update_config(config, self._task_path, info)
        if defaults := config.get("task", "use", fallback=None):
            self._load_config(self._resolve_defaults_config(defaults), False)

    def _read_config(self, config: ConfigParser, path: str) -> bool:
        try:
            res = config.read(path)
        except DuplicateSectionError as e:
            raise TaskConfigParsingError(path, f"Duplicate section [{e.section}]")
        except DuplicateOptionError as e:
            raise TaskConfigParsingError(
                path, f"Duplicate key '{e.option}' in section [{e.section}]"
            )
        except MissingSectionHeaderError as e:
            raise TaskConfigParsingError(path, f"Missing section header")
        return len(res) > 0

    def _resolve_defaults_config(self, name: str):
        def load_from_path(path: str) -> str:
            configs_folder = os.path.join(path, "configs")
            if not os.path.exists(configs_folder):
                raise TaskConfigError(f"'{configs_folder}' does not exist.")
            defaults = os.path.join(configs_folder, name)
            if not os.path.exists(defaults):
                raise TaskConfigError(
                    f"Config '{name}' does not exist in: '{configs_folder}'"
                )
            return defaults

        if name.startswith("@"):
            if (default := V2_DEFAULTS.get(name.removeprefix("@"))) is None:
                raise TaskConfigError(f"Unknown special task_type: '{name}'")
            return default

        if self._pisek_directory is not None:
            return load_from_path(os.path.join(self._task_path, self._pisek_directory))

        current_path = os.path.abspath(self._task_path)
        while current_path:
            if os.path.exists(os.path.join(current_path, ".git")):
                return load_from_path(os.path.join(current_path, "pisek"))
            step_up = os.path.abspath(os.path.join(current_path, ".."))
            if step_up == current_path:
                break  # topmost directory
            if os.stat(step_up).st_uid != os.stat(current_path).st_uid:
                break  # other user's directory
            current_path = step_up

        return name

    def get(self, section: str, key: str | None) -> ConfigValue:
        return self.get_from_candidates([(section, key)])

    def get_from_candidates(
        self, candidates: Iterable[tuple[str, str | None]]
    ) -> ConfigValue:
        unset: bool = False
        for section, key in candidates:
            for config_path, config in zip(self._config_paths, self._configs):
                config_path = os.path.basename(config_path)
                if key is None:
                    value = section if section in config else None
                else:
                    value = config.get(section, key, fallback=None)

                if value == "!unset":
                    unset = True
                    break
                elif value is not None:
                    result = ConfigValue(value, config_path, section, key)
                    self.loaded_values[result] = None
                    return result
            if unset:
                break

        def msg(section_key: tuple[str, str | None]) -> str:
            section, key = section_key
            if key is None:
                return f"Section [{section}]"
            else:
                return f"Key '{key}' in section [{section}]"

        candidates_str = " or\n".join(map(msg, candidates))
        raise TaskConfigError(f"Unset value for:\n{tab(candidates_str)}")

    def get_regex(self, section: str, regex_key: str) -> dict[str, ConfigValue]:
        return self.get_regex_from_candidates([(section, regex_key)])

    def get_regex_from_candidates(
        self, candidates: Iterable[tuple[str, str]]
    ) -> dict[str, ConfigValue]:
        found: dict[str, ConfigValue] = {}
        for section, regex_key in candidates:
            for config_path, config in zip(self._config_paths, self._configs):
                if section not in config:
                    continue

                config_path = os.path.basename(config_path)
                for key, value in config[section].items():
                    if re.fullmatch(regex_key, key) and key not in found:
                        found[key] = ConfigValue(value, config_path, section, key)

        found = {key: val for key, val in found.items() if val != "!unset"}
        self.loaded_values |= {val: None for val in found.values()}
        return found

    def sections(self) -> list[ConfigValue]:
        sections = {
            section: ConfigValue(section, path, section, None)
            for config, path in reversed(list(zip(self._configs, self._config_paths)))
            for section in config.sections()
        }  # We need to use dictionary here because order matters
        return list(sections.values())

    def check_unused(self) -> None:
        """
        Check for unused sections or keys.

        Raises
        ------
        TaskConfigError
            If unused sections or keys are present.
        """
        self._config_helper = ConfigKeysHelper()
        for config_i, config in enumerate(self._configs):
            config_basename = os.path.basename(self._config_paths[config_i])
            for section in config.sections():
                dist, r_section = self._config_helper.find_section(section)
                if dist < 1.0:
                    raise TaskConfigError(
                        f"Unexpected section [{section}] in {config_basename}. "
                        f"(Did you mean [{r_section}]?)"
                    )
                for key in config[section].keys():
                    dist, r_section, r_key = self._config_helper.find_key(
                        section, key, self, config_i > 0
                    )
                    if dist < 1.0:
                        raise TaskConfigError(
                            f"Unexpected key '{key}' in section [{section}] in {config_basename}. "
                            f"(Did you mean '{r_key}'"
                            + (
                                f" in section [{r_section}]"
                                if section != r_section
                                else ""
                            )
                            + "?)"
                        )

    def check_all(self) -> None:
        self.check_unused()

    def check_todos(self) -> bool:
        """Check whether lowest config contains TODO in comments."""
        with open(self._config_paths[0]) as config:
            for line in config:
                if "#" in line and "TODO" in line.split("#")[1]:
                    return True

        return False



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random
from typing import Any

from pisek.utils.paths import InputPath, OutputPath
from pisek.jobs.jobs import Job
from pisek.task_jobs.task_manager import TaskJobManager, SOLUTION_MAN_CODE
from pisek.task_jobs.checker.chaos_monkey import (
    Invalidate,
    Incomplete,
    BlankLine,
    ChaosMonkey,
    TrailingString,
)

from pisek.task_jobs.solution.solution_result import Verdict
from pisek.task_jobs.checker.checker_base import RunChecker
from pisek.task_jobs.checker.cms_judge import RunCMSJudge
from pisek.task_jobs.checker.checker import checker_job


class FuzzingManager(TaskJobManager):
    """Manager that prepares and test judge."""

    def __init__(self) -> None:
        super().__init__("Fuzz judge")

    def _get_jobs(self) -> list[Job]:
        jobs: list[Job] = []

        primary_sol = self.prerequisites_results[
            SOLUTION_MAN_CODE + self._env.config.primary_solution
        ]

        self._inputs: dict[str, tuple[list[int], int | None]] = primary_sol["inputs"]
        testcases: list[tuple[InputPath, OutputPath]] = []
        for inp, res in primary_sol["results"].items():
            testcases.append((inp, res.solution_rr.stdout_file.to_sanitized_output()))

        jt = self._env.config.checks.fuzzing_thoroughness
        CHOICE_JOBS: list[tuple[type[Invalidate], int]] = [
            (Incomplete, jt // 10),
            (BlankLine, jt // 10),
            (ChaosMonkey, 4 * jt // 5),
        ]
        ALL_JOBS: list[type[Invalidate]] = [TrailingString]

        total_times = sum(map(lambda x: x[1], CHOICE_JOBS)) + len(ALL_JOBS) * len(
            testcases
        )
        rand_gen = random.Random(4)  # Reproducibility!
        seeds = rand_gen.sample(range(0, 16**8), total_times)

        if self._env.config.checks.fuzzing_thoroughness:
            for job, times in CHOICE_JOBS:
                for _ in range(times):
                    seed = seeds.pop()
                    inp, out = rand_gen.choice(testcases)
                    jobs += self._fuzz_jobs(job, inp, out, seed, rand_gen)

        if self._env.config.checks.judge_rejects_trailing_string:
            for inp, out in testcases:
                jobs += self._fuzz_jobs(
                    TrailingString,
                    inp,
                    out,
                    seeds.pop(),
                    rand_gen,
                    Verdict.wrong_answer,
                )

        return jobs

    def _fuzz_jobs(
        self,
        job: type[Invalidate],
        inp: InputPath,
        out: OutputPath,
        seed: int,
        rand_gen: random.Random,
        expected_verdict: Verdict | None = None,
    ):
        inv_out = out.to_fuzzing(job.__name__.lower(), seed)
        jobs = [
            invalidate := job(self._env, out, inv_out, seed),  # type: ignore[arg-type, call-arg]
            run_judge := checker_job(
                inp,
                inv_out,
                out,
                rand_gen.choice(self._inputs[inp.name][0]),
                self._inputs[inp.name][1],
                expected_verdict,
                self._env,
            ),
        ]
        run_judge.add_prerequisite(invalidate)
        return jobs

    def _compute_result(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        result["checker_outs"] = set()
        for job in self.jobs:
            if isinstance(job, RunChecker):
                if isinstance(job, RunCMSJudge):
                    result["checker_outs"].add(job.points_file)
                result["checker_outs"].add(job.checker_log_file)

        return result



# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Iterator, Any, Optional
from cms.db.task import Task, Dataset, Manager
from cms.db.filecacher import FileCacher
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound
from os import path, listdir
import re
import datetime
from typing import Callable, TypeVar, assert_never

from pisek.user_errors import NotSupported
from pisek.cms.testcase import create_testcase
from pisek.env.env import Env
from pisek.config.task_config import TaskConfig
from pisek.config.config_types import JudgeType, OutCheck, TaskType, DataFormat
from pisek.utils.paths import TaskPath, InputPath, BUILD_DIR

T = TypeVar("T")


def check_key(name: str, value: T, condition: Callable[[T], bool]):
    if not condition(value):
        raise NotSupported(f"Cannot import task with {name}={value}")


def create_dataset(
    session: Session,
    env: Env,
    task: Task,
    testcases: list[InputPath],
    description: Optional[str],
    time_limit: Optional[float],
    autojudge: bool = True,
) -> Dataset:
    if description is None:
        description = create_description()

    config = env.config

    check_key(
        "out_check",
        config.tests.out_check,
        lambda v: v in (OutCheck.diff, OutCheck.judge, OutCheck.tokens),
    )
    if config.tests.out_check == OutCheck.tokens:
        check_key(
            "tokens_ignore_case", config.tests.tokens_ignore_case, lambda v: not v
        )
        check_key(
            "tokens_ignore_newlines",
            config.tests.tokens_ignore_newlines,
            lambda v: not v,
        )
        check_key(
            "tokens_float_abs_error",
            config.tests.tokens_float_abs_error,
            lambda v: v is None,
        )
        check_key(
            "tokens_float_rel_error",
            config.tests.tokens_float_rel_error,
            lambda v: v is None,
        )
    if (
        config.tests.out_check == OutCheck.judge
        and config.task.task_type == TaskType.batch
    ):
        check_key(
            "judge_type", config.tests.judge_type, lambda t: t == JudgeType.cms_batch
        )
    if (
        config.tests.out_check == OutCheck.judge
        and config.task.task_type == TaskType.interactive
    ):
        check_key(
            "judge_type",
            config.tests.judge_type,
            lambda t: t == JudgeType.cms_communication,
        )

    if config.task.task_type == TaskType.batch:
        check_key(
            "out_format",
            config.tests.out_format,
            lambda t: t in (DataFormat.strict_text, DataFormat.binary),
        )

    score_params = get_group_score_parameters(config)

    task_type: str
    task_params: Any

    if config.task.task_type == TaskType.batch:
        task_type = "Batch"
        task_params = (
            "grader" if config.cms.stubs else "alone",
            ("", ""),
            "comparator" if config.tests.out_check == OutCheck.judge else "diff",
        )
    elif config.task.task_type == TaskType.interactive:
        task_type = "Communication"
        task_params = (1, "stub" if config.cms.stubs else "alone", "std_io")
    else:
        assert_never(config.task.task_type)

    dataset = Dataset(
        description=description,
        autojudge=autojudge,
        task_type=task_type,
        task_type_parameters=task_params,
        score_type="GroupMin",
        score_type_parameters=score_params,
        time_limit=time_limit if time_limit is not None else config.cms.time_limit,
        memory_limit=config.cms.mem_limit * 1024 * 1024,
        task=task,
    )

    session.add(dataset)

    files = FileCacher()

    outputs_needed = (
        config.task.task_type == TaskType.batch and config.tests.judge_needs_out
    )

    for input_ in testcases:
        name = input_.name.removesuffix(".in")
        output: TaskPath | None = None

        if outputs_needed:
            output = input_.to_output()

            if not path.exists(output.path):
                output = TaskPath.data_path(config.primary_solution, output.name)

        create_testcase(session, files, dataset, name, input_, output)

    add_judge(session, files, env, dataset)
    add_stubs(session, files, env, dataset)
    add_headers(session, files, env, dataset)

    return dataset


def get_group_score_parameters(config: TaskConfig) -> list[tuple[int, str]]:
    params = []

    for subtask in config.test_sections.values():
        globs = map(strip_input_extention, subtask.all_globs)
        # CMS supports only relative points therefore we convert 'unscored' to 0
        params.append((subtask.max_points, globs_to_regex(globs)))

    return params


def strip_input_extention(file: str) -> str:
    if not file.endswith(".in"):
        raise RuntimeError(f"Input file {file} does not have a .in extention")

    return file[:-3]


def glob_char_to_regex(c: str) -> str:
    if c == "?":
        return "."
    elif c == "*":
        return ".*"
    else:
        return re.escape(c)


def globs_to_regex(globs: Iterator[str]) -> str:
    patterns = []

    for glob in globs:
        pattern = "".join(map(glob_char_to_regex, glob))
        patterns.append(f"({pattern})")

    return f"^{'|'.join(patterns)}$"


def add_judge(session: Session, files: FileCacher, env: Env, dataset: Dataset):
    config: TaskConfig = env.config

    if config.tests.out_check != OutCheck.judge:
        return

    assert config.tests.out_judge is not None

    run_section = config.tests.out_judge
    judge_path = TaskPath(BUILD_DIR, run_section.exec.path).path

    if path.isdir(judge_path):
        run_path = path.join(judge_path, "run")

        for file in listdir(judge_path):
            file = path.join(judge_path, file)
            assert path.realpath(file) == path.realpath(
                run_path
            ), f"{judge_path} contains {file}, which is not run. Multi-file judges are not supported."

        judge_path = run_path

    if config.task.task_type == TaskType.batch:
        judge_name = "checker"
    elif config.task.task_type == TaskType.interactive:
        judge_name = "manager"
    else:
        assert_never(config.task.task_type)

    judge = files.put_file_from_path(judge_path, f"{judge_name} for {config.cms.name}")
    session.add(Manager(dataset=dataset, filename=judge_name, digest=judge))


MISSING_STUB_ERROR = "Language not supported"
ERROR_STUBS = {
    ".c": f"#error {MISSING_STUB_ERROR}",
    ".cpp": f"#error {MISSING_STUB_ERROR}",
    ".cs": f"#error {MISSING_STUB_ERROR}",
    ".hs": f'{{-# LANGUAGE DataKinds #-}} import GHC.TypeLits; type Error = TypeError (Text "{MISSING_STUB_ERROR}")',
    ".java": MISSING_STUB_ERROR,
    ".pas": f"{{$Message Fatal '{MISSING_STUB_ERROR}'}}",
    ".php": f"<?php throw new Exception('{MISSING_STUB_ERROR}'); ?>",
    ".py": MISSING_STUB_ERROR,
    ".rs": f'compile_error!("{MISSING_STUB_ERROR}");',
}


def add_stubs(session: Session, files: FileCacher, env: Env, dataset: Dataset):
    config = env.config

    if not config.cms.stubs:
        return

    if config.task.task_type == TaskType.batch:
        stub_basename = "grader"
    elif config.task.task_type == TaskType.interactive:
        stub_basename = "stub"
    else:
        assert_never(config.task.task_type)

    exts = set()
    for stub in config.cms.stubs:
        directory, target_name = path.split(stub.path)
        directory = path.normpath(directory)

        for filename in listdir(directory):
            basename, ext = path.splitext(filename)

            if basename != target_name and filename != target_name:
                continue

            stub = files.put_file_from_path(
                path.join(directory, filename),
                f"{stub_basename}{ext} for {config.cms.name}",
            )
            session.add(
                Manager(dataset=dataset, filename=f"{stub_basename}{ext}", digest=stub)
            )

            if ext in exts:
                raise RuntimeError(f"Multiple stubs with extension '{ext}'")
            exts.add(ext)

    for ext, content in ERROR_STUBS.items():
        if ext in exts:
            continue

        stub = files.put_file_content(
            content.encode(), f"{stub_basename}{ext} for {config.cms.name}"
        )
        session.add(
            Manager(dataset=dataset, filename=f"{stub_basename}{ext}", digest=stub)
        )


def add_headers(session: Session, files: FileCacher, env: Env, dataset: Dataset):
    config = env.config

    for header in config.cms.headers:
        name = header.name
        header = files.put_file_from_path(
            header.path, f"Header {name} for {config.cms.name}"
        )
        session.add(Manager(dataset=dataset, filename=name, digest=header))


def get_only_dataset(session: Session, task: Task) -> Dataset:
    datasets = session.query(Dataset).filter(Dataset.task == task).all()

    if len(datasets) >= 2:
        raise RuntimeError(
            f"The task has multiple datasets: {', '.join(sorted(d.description for d in datasets))}"
        )
    else:
        return datasets[0]


def get_dataset_by_description(
    session: Session, task: Task, description: str
) -> Dataset:
    try:
        return (
            session.query(Dataset)
            .filter(Dataset.task == task)
            .filter(Dataset.description == description)
            .one()
        )
    except NoResultFound as e:
        raise RuntimeError(
            f'The task has no dataset with the description "{description}"'
        ) from e


def create_description() -> str:
    date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return date



from pisek.jobs.jobs import Job
from pisek.task_jobs.task_manager import TaskJobManager
from pisek.task_jobs.data.data_manager import TEST_SEED


class EmptyLineManager(TaskJobManager):
    def __init__(self) -> None:
        super().__init__("Empty line")

    def _get_jobs(self) -> list[Job]:
        return []

    def get_status(self):
        return ""


class TestsHeaderManager(TaskJobManager):
    def __init__(self) -> None:
        super().__init__("Tests headers")

    def _get_jobs(self) -> list[Job]:
        return []

    def get_status(self):
        test_header = " " * len(self._solution_header_verbosity0("", 0, 0))
        for num, test in self._env.config.test_sections.items():
            length = sum(
                ti.repeat
                for ti in self._test_testcases(test)
                if test.new_in_test(ti.input_path(TEST_SEED).name)
            )
            if test.direct_predecessors:
                length += 3

            name = test.name
            if len(name) > length:
                name = str(num)
            if len(name) > length:
                name = ""

            test_header += "[" + name.center(length) + "]"
        return test_header



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pisek.jobs.jobs import PipelineItemFailure
from pisek.env.env import Env
from pisek.utils.paths import TaskPath, IInputPath, IOutputPath
from pisek.task_jobs.task_job import TaskJob


class DataJob(TaskJob):
    def __init__(self, env: Env, name: str, data: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env,
            name=name,
            **kwargs,
        )
        self.data = data


class LinkData(DataJob):
    """Link data to into dest folder."""

    def __init__(self, env: Env, data: TaskPath, dest: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env, name=f"Link {data:p} -> {dest:p}", data=data, **kwargs
        )
        self.dest = dest

    def _run(self):
        self._link_file(self.data, self.dest, overwrite=True)


class SymlinkData(DataJob):
    """Link data to into dest folder."""

    def __init__(self, env: Env, data: TaskPath, dest: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env, name=f"Symlink {data:p} -> {dest:p}", data=data, **kwargs
        )
        self.dest = dest

    def _run(self):
        self._symlink_file(self.data, self.dest, overwrite=True)


MB = 1024 * 1024


class InputSmall(DataJob):
    """Checks that input is small enough to download."""

    def __init__(self, env: Env, input_: IInputPath, **kwargs) -> None:
        super().__init__(
            env=env,
            name=f"Input {input_:p} is smaller than {env.config.limits.input_max_size}MB",
            data=input_,
            **kwargs,
        )

    def _run(self):
        max_size = self._env.config.limits.input_max_size
        if (sz := self._file_size(self.data)) > max_size * MB:
            raise PipelineItemFailure(
                f"Input {self.data:p} is bigger than {max_size}MB: {(sz+MB-1)//MB}MB"
            )


class OutputSmall(DataJob):
    """Checks that output is small enough to upload."""

    def __init__(self, env: Env, output: IOutputPath, **kwargs) -> None:
        super().__init__(
            env=env,
            name=f"Output {output:p} is smaller than {env.config.limits.output_max_size}MB",
            data=output,
            **kwargs,
        )

    def _run(self):
        max_size = self._env.config.limits.output_max_size
        if (sz := self._file_size(self.data)) > max_size * MB:
            raise PipelineItemFailure(
                f"Output {self.data} is bigger than {max_size}MB: {(sz+MB-1)//MB}MB"
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, Future, wait

from pisek.env.env import Env
from pisek.jobs.jobs import State, Job, JobManager
from pisek.jobs.cache import Cache
from pisek.jobs.reporting import Reporter, CommandLineReporter


class JobPipeline(ABC):
    """Runs given Jobs and JobManagers according to their prerequisites."""

    @abstractmethod
    def __init__(self) -> None:
        self.job_managers: list[JobManager] = []
        self._tmp_lines: int = 0
        self.all_accessed_files: set[str] = set()

    def run_jobs(self, cache: Cache | None, env: Env) -> bool:
        self._reporter: Reporter = CommandLineReporter(env, self.job_managers)

        self._futures: dict[Future, Job] = {}
        self._queue: list[Job] = []
        self._envs: list[Env] = [env.model_copy(deep=True) for _ in range(env.jobs)]

        for job_man in self.job_managers:
            job_man.set_env(env)

        with ThreadPoolExecutor(max_workers=env.jobs) as self._thread_pool:
            self._update(cache, env)
            while not all(man.state.finished() for man in self.job_managers):
                done, _ = wait(
                    self._futures, timeout=0.1, return_when="FIRST_COMPLETED"
                )

                for future in done:
                    exception = future.exception()
                    if exception is not None:
                        raise exception

                    job = self._futures[future]
                    del self._futures[future]
                    self._envs.append(job._env)
                    self._finalize_job(job, cache)

                if not self._update(cache, env):
                    break

                self._reporter.update(list(self._futures.values()))

            for job in self._futures.values():
                if job.state == State.succeeded:
                    self._finalize_job(job, cache)
                else:
                    job.cancel()
            self._reporter.update([])

        if cache is not None:
            cache.export()  # Save last version of cache

        return any(man.state == State.failed for man in self.job_managers)

    def _update(self, cache: Cache | None, env: Env) -> bool:
        """Updates currently running JobManagers and Jobs, runs new ones
        and returns if we should continue in testing."""

        # Start new managers
        for manager in self.job_managers:
            if manager.state == State.in_queue and manager.prerequisites == 0:
                self._queue.extend(manager.create_jobs())
                if manager.any_failed():
                    manager.finalize()
                    self._reporter.report_manager(manager)
                    if not env.full:
                        return False
                break  # We don't want to start many managers at once because that can lead to UI freeze

        # Process new jobs
        new_queue: list[Job] = []
        to_run: list[tuple[Job, Env]] = []
        for job in self._queue:
            if len(self._envs) > 0 and job.prerequisites == 0:
                job.prepare(cache)
                if job.state.finished():
                    self._finalize_job(job, cache)
                elif job.state:
                    to_run.append((job, self._envs.pop()))
            else:
                new_queue.append(job)

        self._queue = new_queue

        # Update managers
        for manager in self.job_managers:
            if manager.state == State.running:
                manager.update()
                if manager.ready() or manager.any_failed():
                    manager.finalize()
                    self._reporter.report_manager(manager)
                if manager.any_failed() and not env.full:
                    return False

        # Start new jobs
        for job, env in to_run:
            if job.state == State.in_queue:
                self._futures[self._thread_pool.submit(job.run, env)] = job
            else:
                self._envs.append(env)

        return True

    def _finalize_job(self, job: Job, cache: Cache | None) -> None:
        job.finalize(cache)
        self.all_accessed_files |= job.accessed_files
        self._reporter.report_job(job)



import unittest

from util import TestFixtureVariant


class TestCmsBatch(TestFixtureVariant):
    @property
    def fixture_path(self) -> str:
        return "examples/cms-batch/"


class TestOpendata(TestFixtureVariant):
    @property
    def time_limit(self) -> float | None:
        return 1.0

    @property
    def fixture_path(self) -> str:
        return "examples/opendata/"


if __name__ == "__main__":
    unittest.main(verbosity=2)



# Edited version of https://github.com/squidfunk/mkdocs-material/blob/master/src/overrides/hooks/shortcodes.py

from __future__ import annotations

import re

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.files import Files
from mkdocs.structure.pages import Page
from re import Match

# -----------------------------------------------------------------------------
# Hooks
# -----------------------------------------------------------------------------


# @todo
def on_page_markdown(markdown: str, *, page: Page, config: MkDocsConfig, files: Files):

    # Replace callback
    def replace(match: Match):
        type, args = match.groups()
        args = args.strip()
        if type == "version":
            return _badge_for_version(args)
        elif type == "flag":
            return flag(args)
        elif type == "type":
            return _badge_for_type(args)

        elif type == "default":
            return _badge_for_default(args)
        elif type == "default-empty":
            return _badge_for_default_empty()

        # Otherwise, raise an error
        raise RuntimeError(f"Unknown shortcode: {type}")

    # Find and replace all external asset URLs in current page
    return re.sub(r"<!-- md:([\w-]+)(.*?) -->", replace, markdown, flags=re.I | re.M)


# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------


# Create a flag of a specific type
def flag(args: str):
    type, *_ = args.split(" ", 1)
    if type == "experimental":
        return _badge_for_experimental()
    elif type == "required":
        return _badge_for_required()
    elif type == "required-applicable":
        return _badge_for_required_applicable()
    elif type == "customization":
        return _badge_for_customization()
    raise RuntimeError(f"Unknown type: {type}")


# -----------------------------------------------------------------------------


# Create badge
def _badge(icon: str, text: str = "", type: str = ""):
    classes = f"mdx-badge mdx-badge--{type}" if type else "mdx-badge"
    return "".join(
        [
            f'<span class="{classes}">',
            *([f'<span class="mdx-badge__icon">{icon}</span>'] if icon else []),
            *([f'<span class="mdx-badge__text">{text}</span>'] if text else []),
            f"</span>",
        ]
    )


# Create badge for version
def _badge_for_version(text: str):
    icon = "material-tag-outline"
    return _badge(
        icon=f":{icon}:{{ title='Minimum version' }}",
        text=(
            f"[{text}](https://github.com/piskoviste/pisek/releases/tag/v{text})"
            if text
            else ""
        ),
    )


# Create badge for type
def _badge_for_type(text: str):
    icon = "fontawesome-solid-triangle-circle-square"
    return _badge(
        icon=f":{icon}:{{ title='Type' }}",
        text=(f"[{text}](#value-types)" if text else ""),
    )


# Create badge for default value
def _badge_for_default(text: str):
    icon = "material-water"
    return _badge(icon=f":{icon}:{{ title='Default value' }}", text=text)


# Create badge for empty default value
def _badge_for_default_empty():
    icon = "material-water-outline"
    return _badge(icon=f":{icon}:{{ title='Default value is empty' }}")


# Create badge for required value flag
def _badge_for_required():
    icon = "material-alert"
    return _badge(icon=f":{icon}:{{ title='Required value' }}")


# Create badge for required value flag
def _badge_for_required_applicable():
    icon = "material-alert-plus"
    return _badge(icon=f":{icon}:{{ title='Required value if applicable' }}")


# Create badge for customization flag
def _badge_for_customization():
    icon = "material-brush-variant"
    return _badge(icon=f":{icon}:{{ title='Customization' }}")


# Create badge for experimental flag
def _badge_for_experimental():
    icon = "material-flask-outline"
    return _badge(icon=f":{icon}:{{ title='Experimental' }}")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fnmatch
from functools import cached_property
import os
from pydantic_core import PydanticCustomError, ErrorDetails
from pydantic import (
    Field,
    computed_field,
    field_validator,
    BeforeValidator,
    TypeAdapter,
    ValidationError,
    ValidationInfo,
    model_validator,
)
import re
from typing import Any, Annotated, ClassVar, Mapping, Optional, TypeVar, Union

from pisek.utils.paths import TaskPath
from pisek.utils.text import tab
from pisek.utils.text import warn
from pisek.env.base_env import BaseEnv
from pisek.config.config_hierarchy import ConfigValue, TaskConfigError, ConfigHierarchy
from pisek.config.config_types import (
    TaskType,
    GenType,
    ValidatorType,
    OutCheck,
    JudgeType,
    ShuffleMode,
    DataFormat,
    TestPoints,
    ProgramRole,
    BuildStrategyName,
    CMSFeedbackLevel,
    CMSScoreMode,
)
from pisek.env.context import init_context
from pisek.task_jobs.solution.solution_result import TEST_SPEC
from pisek.task_jobs.builder.strategies import ALL_STRATEGIES


T = TypeVar("T")
Maybe = Annotated[T | None, BeforeValidator(lambda t: t or None)]


MaybeInt = Annotated[
    Optional[int], BeforeValidator(lambda i: (None if i == "X" else i))
]
ListStr = Annotated[list[str], BeforeValidator(lambda s: s.split())]

TaskPathFromStr = Annotated[TaskPath, BeforeValidator(lambda s: TaskPath(s))]
OptionalTaskPathFromStr = Annotated[
    Optional[TaskPath], BeforeValidator(lambda s: TaskPath(s) if s else None)
]
ListTaskPathFromStr = Annotated[
    list[TaskPath], BeforeValidator(lambda s: [TaskPath(p) for p in s.split()])
]

MISSING_VALIDATION_CONTEXT = "Missing validation context."

ValuesDict = dict[str, Union[str, "ValuesDict", dict[Any, "ValuesDict"]]]
ConfigValuesDict = Mapping[
    str, Union[ConfigValue, "ConfigValuesDict", dict[Any, "ConfigValuesDict"]]
]


def _to_values(config_values_dict: ConfigValuesDict) -> ValuesDict:
    def convert(what: ConfigValue | Mapping) -> str | dict:
        if isinstance(what, ConfigValue):
            return what.value
        else:
            return {key: convert(val) for key, val in what.items()}

    return {key: convert(val) for key, val in config_values_dict.items()}


def _validate_program_name(key: str, value: str) -> str:
    error_key = key.lower().replace(" ", "_")
    for banned_char in "[]":
        if banned_char in value:
            raise PydanticCustomError(
                f"invalid_{error_key}",
                f"{key.capitalize()} must not contain '{banned_char}'",
            )
    if value.startswith("_"):
        raise PydanticCustomError(
            f"invalid_{error_key}",
            f"{key.capitalize()} must not start with '_'",
        )
    return value


class TaskConfig(BaseEnv):
    """Configuration of task loaded from config file."""

    task: "TaskSection"
    tests: "TestsSection"
    test_sections: dict[int, "TestSection"]

    solutions: dict[str, "SolutionSection"]

    solution_time_limit: float = Field(ge=0)  # Needed for visualization

    limits: "LimitsSection"

    cms: "CMSSection"

    checks: "ChecksSection"

    @computed_field  # type: ignore[misc]
    @cached_property
    def total_points(self) -> int:
        return sum(sub.max_points for sub in self.test_sections.values())

    @computed_field  # type: ignore[misc]
    @property
    def tests_count(self) -> int:
        return len(self.test_sections)

    @computed_field  # type: ignore[misc]
    @cached_property
    def input_globs(self) -> list[str]:
        return sum((sub.all_globs for sub in self.test_sections.values()), start=[])

    @computed_field  # type: ignore[misc]
    @property
    def primary_solution(self) -> str:
        if len(self.solutions) == 0:
            raise RuntimeError("No solutions exist.")
        else:
            return [name for name, sol in self.solutions.items() if sol.primary][0]

    @computed_field  # type: ignore[misc]
    @cached_property
    def max_solution_time_limit(self) -> float:
        return max(
            (solution.run.time_limit for solution in self.solutions.values()), default=0
        )

    def get_solution_by_run(self, run: str) -> Optional[str]:
        sources = (name for name, sol in self.solutions.items() if sol.run.name == run)
        return next(sources, None)

    def __init__(self, **kwargs):
        value = {"test_count": max(kwargs["test_sections"]) + 1}

        with init_context(value):
            super().__init__(**kwargs)

    @staticmethod
    def load_dict(configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, Any] = {}

        args["task"] = TaskSection.load_dict(configs)
        args["tests"] = TestsSection.load_dict(args["task"]["task_type"].value, configs)

        section_names = configs.sections()

        PROGRAMS = [
            (ProgramRole.gen, "in_gen"),
            (ProgramRole.validator, "validator"),
            (ProgramRole.judge, "out_judge"),
        ]
        for t, program in PROGRAMS:
            if args["tests"][program].value:
                args["tests"][program] = RunSection.load_dict(
                    t, args["tests"][program], configs
                )

        # Load tests
        args["test_sections"] = test_sec = {}
        # Sort so tests.keys() returns tests in sorted order
        for section in sorted(section_names, key=lambda cv: cv.value):
            section_name = section.value
            if m := re.fullmatch(r"test(\d{2})", section_name):
                num = m[1]
                test_sec[int(num)] = TestSection.load_dict(
                    ConfigValue(str(int(num)), section.config, section.section, None),
                    configs,
                )

        args["solutions"] = solutions = {}
        for section in section_names:
            if m := re.fullmatch(r"solution_(.+)", section.value):
                solutions[m[1]] = SolutionSection.load_dict(
                    ConfigValue(m[1], section.config, section.section, None), configs
                )

        args["limits"] = LimitsSection.load_dict(configs)
        args["cms"] = CMSSection.load_dict(configs)
        args["checks"] = ChecksSection.load_dict(configs)

        args["solution_time_limit"] = configs.get_from_candidates(
            [("run_solution", "time_limit"), ("run", "time_limit")]
        )

        return args

    @model_validator(mode="after")
    def validate_model(self):
        if (
            self.task.task_type == TaskType.interactive
            and self.tests.out_check != OutCheck.judge
        ):
            raise PydanticCustomError(
                "interactive_must_have_judge",
                "For interactive task 'out_check' must be 'judge'",
                {"task_type": self.task.task_type, "out_check": self.tests.out_check},
            )

        JUDGE_TYPES = {
            TaskType.batch: [
                None,
                JudgeType.opendata_v1,
                JudgeType.opendata_v2,
                JudgeType.cms_batch,
            ],
            TaskType.interactive: [JudgeType.cms_communication],
        }

        if self.tests.judge_type not in JUDGE_TYPES[self.task.task_type]:
            raise PydanticCustomError(
                "task_judge_type_mismatch",
                f"'{self.tests.judge_type}' judge for '{self.task.task_type}' task is not allowed",
                {"task_type": self.task.task_type, "judge_type": self.tests.judge_type},
            )

        primary = [name for name, sol in self.solutions.items() if sol.primary]
        if len(primary) > 1:
            raise PydanticCustomError(
                "multiple_primary_solutions",
                "Multiple primary solutions",
                {"primary_solutions": primary},
            )
        if len(self.solutions) > 0 and len(primary) == 0:
            raise PydanticCustomError(
                "no_primary_solution",
                "No primary solution set",
                {},
            )

        for i in range(len(self.test_sections)):
            if i not in self.test_sections:
                raise PydanticCustomError(
                    "missing_test",
                    f"Missing section [test{i:02}]",
                    {},
                )

        self._compute_predecessors()
        return self

    def _compute_predecessors(self) -> None:
        visited = set()
        computed = set()

        def compute_test(num: int) -> list[int]:
            test = self.test_sections[num]
            if num in computed:
                return test.all_predecessors
            elif num in visited:
                raise PydanticCustomError(
                    "cyclic_predecessor_tests", "Cyclic predecessor tests", {}
                )

            visited.add(num)
            all_predecessors = sum(
                (compute_test(p) for p in test.direct_predecessors),
                start=test.direct_predecessors,
            )

            def normalize_list(l):
                return list(sorted(set(l)))

            test.all_predecessors = normalize_list(all_predecessors)
            test.prev_globs = normalize_list(
                sum(
                    (self.test_sections[p].in_globs for p in test.all_predecessors),
                    start=[],
                )
            )
            test.all_globs = normalize_list(test.prev_globs + test.in_globs)
            computed.add(num)

            return test.all_predecessors

        for i in range(self.tests_count):
            compute_test(i)


class TaskSection(BaseEnv):
    _section: str = "task"
    task_type: TaskType
    score_precision: int = Field(ge=0)

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, ConfigValue] = {
            key: configs.get("task", key) for key in cls.model_fields
        }
        return {"_section": configs.get("task", None), **args}


class TestsSection(BaseEnv):
    _section: str = "tests"

    static_subdir: TaskPathFromStr
    in_gen: Maybe["RunSection"]
    gen_type: Maybe[GenType]

    validator: Maybe["RunSection"]
    validator_type: Maybe[ValidatorType]

    out_check: OutCheck
    out_judge: Maybe["RunSection"]
    judge_type: Maybe[JudgeType]
    judge_needs_in: bool | None
    judge_needs_out: bool | None
    tokens_ignore_newlines: bool | None
    tokens_ignore_case: bool | None
    tokens_float_rel_error: Maybe[float]
    tokens_float_abs_error: Maybe[float]
    shuffle_mode: Maybe[ShuffleMode]
    shuffle_ignore_case: bool | None

    in_format: DataFormat
    out_format: DataFormat

    @staticmethod
    def load_dict(task_type: str, configs: ConfigHierarchy) -> ConfigValuesDict:
        GLOBAL_KEYS = [
            "in_gen",
            "validator",
            "out_check",
            "in_format",
            "out_format",
            "static_subdir",
        ]
        PROGRAM_TYPES = [
            ("gen_type", "in_gen"),
            ("validator_type", "validator"),
        ]
        OUT_CHECK_SPECIFIC_KEYS = [
            ((None, "judge"), "out_judge", ""),
            ((None, "judge"), "judge_type", ""),
            ((TaskType.batch, "judge"), "judge_needs_in", "0"),
            ((TaskType.batch, "judge"), "judge_needs_out", "1"),
            ((None, "tokens"), "tokens_ignore_newlines", "0"),
            ((None, "tokens"), "tokens_ignore_case", "0"),
            ((None, "tokens"), "tokens_float_rel_error", ""),
            ((None, "tokens"), "tokens_float_abs_error", ""),
            ((None, "shuffle"), "shuffle_mode", ""),
            ((None, "shuffle"), "shuffle_ignore_case", "0"),
        ]
        args: dict[str, ConfigValue] = {
            key: configs.get("tests", key) for key in GLOBAL_KEYS
        }

        for program_type, program in PROGRAM_TYPES:
            if args[program].value:
                args[program_type] = configs.get("tests", program_type)
            else:
                args[program_type] = ConfigValue.make_internal(
                    "", "tests", program_type
                )

        # Load judge specific keys
        for (task_type_cond, out_check), key, default in OUT_CHECK_SPECIFIC_KEYS:
            if (task_type_cond is None or task_type_cond == task_type) and args[
                "out_check"
            ].value == out_check:
                args[key] = configs.get("tests", key)
            else:
                args[key] = ConfigValue.make_internal(default, "tests", key)

        return {"_section": configs.get("tests", None), **args}

    @model_validator(mode="after")
    def validate_model(self):
        if self.in_gen is not None and self.gen_type is None:
            raise PydanticCustomError(
                "no_gen_type",
                "Missing gen_type",
                {"in_gen": self.in_gen.name, "gen_type": ""},
            )

        if self.validator is not None and self.validator_type is None:
            raise PydanticCustomError(
                "no_validator_type",
                "Missing validator_type",
                {"validator": self.validator.name, "validator_type": ""},
            )

        if (self.tokens_float_abs_error is not None) != (
            self.tokens_float_rel_error is not None
        ):
            raise PydanticCustomError(
                "tokens_errors_must_be_set_together",
                "Both types of floating point error must be set together",
                {
                    "tokens_float_abs_error": self.tokens_float_abs_error,
                    "tokens_float_rel_error": self.tokens_float_rel_error,
                },
            )

        return self


class TestSection(BaseEnv):
    """Configuration of one test (group of testcases)."""

    _section: str
    num: int
    name: str
    points: TestPoints
    in_globs: ListStr
    prev_globs: list[str] = []
    all_globs: list[str] = []
    direct_predecessors: list[int]
    all_predecessors: list[int] = []
    checks_validate: bool
    checks_different_outputs: bool

    @property
    def max_points(self) -> int:
        return 0 if self.points == "unscored" else self.points

    def in_test(self, filename: str) -> bool:
        return any(fnmatch.fnmatch(filename, g) for g in self.all_globs)

    def new_in_test(self, filename: str) -> bool:
        return not any(
            fnmatch.fnmatch(filename, g) for g in self.prev_globs
        ) and self.in_test(filename)

    @staticmethod
    def load_dict(number: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = [
            "name",
            "points",
            "in_globs",
            "predecessors",
            "checks.validate",
            "checks.different_outputs",
        ]
        args: dict[str, Any] = {
            key.replace(".", "_"): configs.get_from_candidates(
                [(number.section, key), ("tests", key)]
            )
            for key in KEYS
        }
        args["direct_predecessors"] = args.pop("predecessors")

        return {"_section": configs.get(number.section, None), "num": number, **args}

    @field_validator("in_globs", mode="after")
    @classmethod
    def validate_globs(cls, value: list[str], info: ValidationInfo) -> list[str]:
        globs = []
        for glob in value:
            if glob == "@ith":
                glob = f"{info.data['num']:02}*.in"
            if not glob.endswith(".in"):
                raise PydanticCustomError(
                    "in_globs_end_in", "In_globs must end with '.in'"
                )
            globs.append(glob)

        return globs

    @field_validator("direct_predecessors", mode="before")
    @classmethod
    def expand_predecessors(cls, value: str, info: ValidationInfo) -> list[str]:
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)
        test_cnt = info.context.get("test_count")
        number = info.data["num"]

        predecessors = []
        for pred in value.split():
            if pred == "@previous":
                if number <= 1:
                    continue
                predecessors.append(number - 1)
            else:
                try:
                    num = int(pred)
                except ValueError:
                    raise PydanticCustomError(
                        "predecessors_must_be_int", "Predecessors must be int"
                    )
                if not 0 <= num < test_cnt:
                    raise PydanticCustomError(
                        "predecessors_must_be_in_range",
                        f"Predecessors must be in range 0, {test_cnt-1}",
                    )
                predecessors.append(num)

        return list(sorted(set(predecessors)))

    @model_validator(mode="after")
    def validate_model(self):
        if self.name == "@auto":
            self.name = f"Test {self.num}"

        return self


class SolutionSection(BaseEnv):
    """Configuration of one solution."""

    _section: str
    name: str
    primary: bool
    run: "RunSection"
    points: MaybeInt
    points_min: MaybeInt
    points_max: MaybeInt
    tests: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def load_dict(cls, name: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = [
            "primary",
            "run",
            "points",
            "points_min",
            "points_max",
            "tests",
        ]
        args = {
            key: configs.get_from_candidates([(name.section, key), ("solutions", key)])
            for key in KEYS
        }

        if args["run"].value == "@auto":
            args["run"] = args["run"].change_value(name.value)

        sol_type = (
            ProgramRole.primary_solution
            if TypeAdapter(bool).validate_strings(args["primary"].value)
            else ProgramRole.secondary_solution
        )

        return {
            "_section": configs.get(name.section, None),
            "name": name,
            **args,
            "run": RunSection.load_dict(sol_type, args["run"], configs),
        }

    @field_validator("name", mode="after")
    @classmethod
    def validate_name(cls, value: str) -> str:
        return _validate_program_name("solution name", value)

    @field_validator("tests", mode="after")
    def validate_tests(cls, value, info: ValidationInfo):
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)
        test_cnt = info.context.get("test_count")
        primary = info.data.get("primary")
        if value == "@auto":
            value = ("1" if primary else "X") * test_cnt
        elif value == "@all":
            value = "1" * test_cnt
        elif value == "@any":
            value = "X" * test_cnt

        if len(value) != test_cnt:
            raise PydanticCustomError(
                "tests_str_invalid_len",
                f"There are {test_cnt} tests but test string has {len(value)} characters",
            )

        for char in value:
            if char not in TEST_SPEC:
                raise PydanticCustomError(
                    "tests_str_invalid_char",
                    f"Not allowed char in test string: {char}. Recognized are {''.join(TEST_SPEC.keys())}",
                )

        if primary and value != "1" * test_cnt:
            raise PydanticCustomError(
                "primary_sol_must_succeed",
                f"Primary solution must have: tests={'1'*test_cnt}",
            )

        return value

    @model_validator(mode="after")
    def validate_model(self):
        for points_limit in ["points_min", "points_max"]:
            if self.points is not None and getattr(self, points_limit) is not None:
                raise PydanticCustomError(
                    "points_double_set",
                    f"Both 'points' and '{points_limit}' are set at once",
                    {"points": self.points, points_limit: getattr(self, points_limit)},
                )

        return self


def get_run_defaults(program_role: ProgramRole, program_name: str) -> list[str]:
    if program_role.is_solution():
        return [
            f"run_solution:{program_name}",
            f"run_{program_role}",
            f"run_solution",
            f"run",
        ]
    else:
        return [
            f"run_{program_role}:{program_name}",
            f"run_{program_role}",
            f"run",
        ]


class RunSection(BaseEnv):
    """Configuration of running an program"""

    _section: str

    program_role: ProgramRole
    name: str
    subdir: str
    build: "BuildSection"
    exec: TaskPathFromStr
    time_limit: float = Field(ge=0)  # [seconds]
    clock_mul: float = Field(ge=0)  # [1]
    clock_min: float = Field(ge=0)  # [seconds]
    mem_limit: int = Field(ge=0)  # [KB]
    process_limit: int = Field(ge=0)  # [1]
    # limit=0 means unlimited
    args: ListStr
    env: dict[str, str]

    @property
    def executable(self) -> TaskPath:
        return TaskPath.executable_path(self.build.program_name, self.exec.path)

    def clock_limit(self, override_time_limit: Optional[float] = None) -> float:
        tl = override_time_limit if override_time_limit is not None else self.time_limit
        if tl == 0:
            return 0
        return max(tl * self.clock_mul, self.clock_min)

    @classmethod
    def load_dict(
        cls, program_role: ProgramRole, name: ConfigValue, configs: ConfigHierarchy
    ) -> ConfigValuesDict:
        default_sections = get_run_defaults(program_role, name.value)

        section_name = configs.get_from_candidates(
            [(section, None) for section in default_sections]
        )
        args = {
            key: configs.get_from_candidates(
                [(section, key) for section in default_sections]
            )
            for key in cls.model_fields
            if key not in ("name", "program_role", "env")
        }
        if args["build"].value == "@auto":
            args["build"] = args["build"].change_value(
                f"{program_role.build_name}:{os.path.join(args['subdir'].value, name.value)}"
            )

        envs = configs.get_regex_from_candidates(
            (section, "env_(.*)") for section in default_sections
        )
        envs = {key.removeprefix("env_"): val for key, val in envs.items()}

        return {
            "_section": section_name,
            "program_role": ConfigValue.make_internal(
                program_role.name, "run", "program_role"
            ),
            "name": name,
            **args,
            "build": BuildSection.load_dict(args["build"], configs),
            "env": envs,
        }

    @field_validator("name", mode="after")
    @classmethod
    def validate_name(cls, value: str) -> str:
        return _validate_program_name("program name", value)

    @field_validator("env", mode="before")
    def interpolate_env(cls, dictionary: dict[str, str]) -> dict[str, str]:
        def raise_err(error_type: str, message: str, key: str):
            raise PydanticCustomError(
                error_type, message, {"_loc": key, f"env_{key}": dictionary[key]}
            )

        def interpolate(key: str, val: str) -> str:
            state: list[str] = [""]
            i = 0
            while i < len(val):
                char = val[i]
                if char == "\\":
                    if i == len(val) - 1:
                        raise_err("env_wrong_escape", "Unterminated escape", key)
                    i += 1
                    state[-1] += val[i]
                elif val[i : i + 2] == "${":
                    state.append("")
                    i += 1
                elif char == "}" and len(state) > 1:
                    var_name = state.pop()
                    if var_name not in os.environ:
                        raise_err(
                            "unknown_variable",
                            f"No environment variable '{var_name}'",
                            key,
                        )
                    state[-1] += os.environ[var_name]
                elif char in "${}":
                    raise_err("unescaped_char", f"Unescaped character '{char}'", key)
                else:
                    state[-1] += char

                i += 1

            if len(state) > 1:
                raise_err(
                    "env_unterminated_interpolation", "Unterminated interpolation", key
                )

            return state[0]

        for key in dictionary:
            dictionary[key] = interpolate(key, dictionary[key])
        return dictionary


class BuildSection(BaseEnv):
    program_names: ClassVar[dict[str, str]] = {}

    _section: str
    section_name: str
    build_type: str
    program_name: str

    sources: ListTaskPathFromStr
    comp_args: ListStr
    extras: ListTaskPathFromStr
    strategy: BuildStrategyName
    entrypoint: str

    headers_c: ListTaskPathFromStr
    extra_sources_c: ListTaskPathFromStr
    headers_cpp: ListTaskPathFromStr
    extra_sources_cpp: ListTaskPathFromStr
    extra_sources_py: ListTaskPathFromStr
    extra_sources_java: ListTaskPathFromStr

    @classmethod
    def load_dict(cls, name: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        program = name
        program_role = ConfigValue("", name.config, name.section, name.key)
        default_sections = [f"build:{program.value}", "build"]
        for pt in ProgramRole:
            prefix = f"{pt.build_name}:"
            if name.value.startswith(prefix):
                program_role, program = name.split(":")
                default_sections = [
                    f"build_{pt.build_name}:{program.value}",
                    f"build_{pt.build_name}",
                    f"build",
                ]
                break

        if (
            program.value in cls.program_names
            and cls.program_names[program.value] != default_sections[0]
        ):
            raise TaskConfigError(
                "Colliding suffixes of build sections not allowed: "
                f"[{default_sections[0]}] and [{cls.program_names[program.value]}]."
            )
        cls.program_names[program.value] = default_sections[0]

        section_name = configs.get_from_candidates(
            [(section, None) for section in default_sections]
        )
        args = {
            key: configs.get_from_candidates(
                [(section, key) for section in default_sections]
            )
            for key in cls.model_fields
            if key not in ("section_name", "build_type", "program_name")
        }

        return {
            "_section": section_name,
            "section_name": ConfigValue.make_internal(
                default_sections[0], default_sections[0], None
            ),
            "build_type": program_role,
            "program_name": program,
            **args,
        }

    @field_validator("program_name", mode="after")
    @classmethod
    def validate_program_name(cls, value: str) -> str:
        return _validate_program_name("program name", value)

    @field_validator("sources", mode="before")
    @classmethod
    def convert_sources(cls, value: str, info: ValidationInfo) -> str:
        if value == "@auto":
            return str(info.data.get("program_name"))
        else:
            return value


class LimitsSection(BaseEnv):
    """Configuration of input and output size limits."""

    _section: str = "limits"

    input_max_size: int
    output_max_size: int

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, ConfigValue] = {
            key: configs.get("limits", key) for key in cls.model_fields
        }
        return {"_section": configs.get("limits", None), **args}


class CMSSection(BaseEnv):
    _section: str = "cms"

    name: Maybe[str]
    title: str
    submission_format: ListStr

    time_limit: float = Field(gt=0)  # [seconds]
    mem_limit: int = Field(gt=0)  # [KB]

    max_submissions: MaybeInt = Field(gt=0)
    min_submission_interval: int = Field(ge=0)  # [seconds]

    score_mode: CMSScoreMode
    feedback_level: CMSFeedbackLevel

    stubs: ListTaskPathFromStr
    headers: ListTaskPathFromStr

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args = {key: configs.get("cms", key) for key in cls.model_fields}

        def get_strategy_union(key: str) -> str:
            all_items = {
                configs.get_from_candidates(
                    [
                        ("build_solution", getattr(strat, key)),
                        ("build", getattr(strat, key)),
                    ]
                ).value
                for strat in ALL_STRATEGIES.values()
                if getattr(strat, key) is not None
            }
            return " ".join(sorted(all_items))

        if args["stubs"].value == "@auto":
            args["stubs"] = args["stubs"].change_value(
                get_strategy_union("extra_sources")
            )
        if args["headers"].value == "@auto":
            args["headers"] = args["headers"].change_value(
                get_strategy_union("extra_nonsources")
            )

        return {"_section": configs.get("cms", None), **args}

    @field_validator("title", mode="before")
    @classmethod
    def convert_title(cls, value: str, info: ValidationInfo) -> str:
        if value == "@name":
            name = info.data.get("name")
            name = "unnamed-task" if name is None else name

            return name
        else:
            return value

    @field_validator("submission_format", mode="after")
    @classmethod
    def convert_format(cls, value: list[str], info: ValidationInfo) -> list[str]:
        name = info.data.get("name")
        name = "unnamed-task" if name is None else name

        return [
            (CMSSection.get_default_file_name(name) if n == "@name" else n)
            for n in value
        ]

    @classmethod
    def get_default_file_name(cls, name: str):
        name = re.sub(r"[^a-zA-Z0-9]+", "_", name)
        return f"{name}.%l"


class ChecksSection(BaseEnv):
    """Configuration of checks for pisek to run."""

    _section: str = "checks"

    solution_for_each_test: bool
    no_unused_inputs: bool
    all_inputs_in_last_test: bool
    generator_respects_seed: bool
    one_input_in_each_nonsample_test: bool
    fuzzing_thoroughness: int
    judge_rejects_trailing_string: bool

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, ConfigValue] = {
            key: configs.get("checks", key) for key in cls.model_fields
        }
        return {"_section": configs.get("checks", None), **args}


def _format_message(err: ErrorDetails) -> str:
    inp = err["input"]
    ctx = err["ctx"] if "ctx" in err else None
    if isinstance(inp, dict) and ctx is not None:
        if ctx == {}:
            return f"{err['msg']}."
        return f"{err['msg']}:\n" + tab(
            "\n".join(
                f"{key}={val}" for key, val in ctx.items() if not key.startswith("_")
            )
        )
    return f"{err['msg']}: '{inp}'"


def _convert_errors(e: ValidationError, config_values: ConfigValuesDict) -> list[str]:
    error_msgs: list[str] = []
    for error in e.errors():
        value: Any = config_values
        for loc in error["loc"]:
            value = value[loc]

        if isinstance(value, ConfigValue):
            location = value.location()
        elif "_section" in value:
            location = value["_section"].location()
        else:
            location = value[error["ctx"]["_loc"]].location()

        error_msgs.append(f"In {location}:\n" + tab(_format_message(error)))
    return error_msgs


def load_config(
    path: str,
    pisek_directory: Optional[str],
    config_filename: str,
    strict: bool = False,
    suppress_warnings: bool = False,
) -> TaskConfig:
    """Loads config from given path."""
    try:
        config_hierarchy = ConfigHierarchy(
            path, not suppress_warnings, pisek_directory, config_filename
        )
        config_values = TaskConfig.load_dict(config_hierarchy)
        config = TaskConfig(**_to_values(config_values))
        config_hierarchy.check_all()
        if config_hierarchy.check_todos() and not suppress_warnings:
            warn("Unsolved TODOs in config.", TaskConfigError, strict)
        return config
    except ValidationError as err:
        raise TaskConfigError(
            "Invalid config:\n\n" + "\n\n".join(_convert_errors(err, config_values))
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Optional

from pisek.user_errors import InvalidArgument
from pisek.utils.text import tab
from pisek.config.task_config import TaskConfig, SolutionSection


class UnknownSolutions(InvalidArgument):
    def __init__(
        self, unknown: list[str], solutions: dict[str, "SolutionSection"]
    ) -> None:
        def format_solution(s: SolutionSection) -> str:
            if s.name == s.run.name:
                return s.name
            else:
                return f"{s.name} (run: {s.run.name})"

        unknown_text = "', '".join(unknown)
        sols_text = "\n".join(map(format_solution, solutions.values()))
        super().__init__(
            f"Unknown solution{'s' if len(unknown) > 1 else ''} '{unknown_text}'. "
            f"Known are:\n{tab(sols_text)}"
        )


def select_solution(config: TaskConfig, solution: str) -> Optional[str]:
    """Tries to find solution by its name/run."""
    if solution in config.solutions:
        return solution

    if (name := config.get_solution_by_run(solution)) is not None:
        return name

    return config.get_solution_by_run(os.path.splitext(solution)[0])


def expand_solutions(config: TaskConfig, solutions: Optional[list[str]]) -> list[str]:
    if solutions is None:
        return list(config.solutions)

    expanded = []
    unknown = []
    for sol in solutions:
        exp = select_solution(config, sol)
        if exp is not None:
            expanded.append(exp)
        else:
            unknown.append(sol)

    if unknown:
        raise UnknownSolutions(unknown, config.solutions)

    return expanded



from colorama import Cursor
from readchar import readkey, key
from typing import Sequence, TypeVar

from pisek.utils.colors import ColorSettings

T = TypeVar("T")


def input_string(message: str) -> str:
    inp = ""
    while not inp:
        inp = input(message).strip()
    return inp


def input_choice(message: str, choices: Sequence[tuple[T, str]], no_jumps: bool) -> T:
    if no_jumps:
        return input_choice_no_jumps(message, choices)

    assert choices

    print(message)

    selected = 0
    while True:
        for i, (_, text) in enumerate(choices):
            full_text = f" {i+1}. {text}"
            if selected == i:
                selector = ColorSettings.colored(">", "cyan")
                print(ColorSettings.colored_back(selector + full_text, "lightblack_ex"))
            else:
                print(" " + full_text)

        k = readkey()
        if k in (key.SPACE, key.ENTER):
            return choices[selected][0]
        elif k in "123456789":
            selected = min(int(k) - 1, len(choices) - 1)
        elif k == key.DOWN:
            selected = (selected + 1) % len(choices)
        elif k == key.UP:
            selected = (selected - 1) % len(choices)

        print(Cursor.UP() * len(choices), end="")


def input_choice_no_jumps(message: str, choices: Sequence[tuple[T, str]]) -> T:
    assert choices

    print(message)

    for i, (_, text) in enumerate(choices):
        full_text = f" {i+1}. {text}"
        print(" " + full_text)

    while True:
        try:
            selected = int(input(f"Enter number from 1 to {len(choices)}: ")) - 1
            if 0 <= selected < len(choices):
                return choices[selected][0]
        except ValueError:
            pass



#!/usr/bin/env python3
# Validator based on Kasiopea's validator template from Pali Madaj.
import sys
import argparse


def main(diff):
    # Here BOUNDS[2] are stricter than what the generator really creates.
    BOUNDS = [(-1e9, 1e9), (-1e9, 1e9)]

    t = read_values(1)[0]
    for ti in range(t):
        read_values(2, *BOUNDS[diff])


# ----------------- Cast nize jsou pomocne funkce, snad nebude treba upravovat -----------------

line_number = 0


def read_line():
    global line_number
    line_number += 1
    return input()


def fail(message):
    global line_number
    print(message + " (na radku {})".format(line_number), file=sys.stderr)
    sys.exit(1)


def read_values(count=None, minimum=None, maximum=None, value_type=int):
    try:
        line = read_line()
    except EOFError:
        fail("Konec souboru.")

    try:
        numbers = list(map(value_type, line.split(" ")))
    except ValueError:
        fail("Hodnota neni typu {}.".format(value_type))

    if count is not None and len(numbers) != count:
        fail("Pocet hodnot na radku byl {} a mel byt {}.".format(len(numbers), count))
    if minimum is not None and any(x < minimum for x in numbers):
        fail("Nejaka hodnota na radku byla mensi nez {}.".format(minimum))
    if maximum is not None and any(x > maximum for x in numbers):
        fail("Nejaka hodnota na radku byla vetsi nez {}.".format(maximum))

    return numbers


def expect_eof():
    koniec = False
    try:
        input()
    except EOFError:
        koniec = True
    if not koniec:
        fail("Soubor pokracuje, ale mel uz skoncit.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Nacte ze stdin vstupy ulohy a zkontroluje jejich spravnost"
    )
    parser.add_argument("test", type=int, help="test (indexovany od 1)")
    args = parser.parse_args()
    main(args.test - 1)
    expect_eof()
    exit(42)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from decimal import Decimal
import logging
from typing import Any

from pisek.utils.paths import IInputPath, IOutputPath
from pisek.env.env import Env
from pisek.config.config_types import ProgramRole
from pisek.config.task_config import RunSection
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    SolutionResult,
    AbsoluteSolutionResult,
    RelativeSolutionResult,
)

from pisek.task_jobs.checker.checker_base import RunBatchChecker

logger = logging.getLogger(__name__)


OPENDATA_NO_SEED = "-"


class RunOpendataJudge(RunBatchChecker):
    """Checks solution output using judge with the opendata interface. (Abstract class)"""

    @property
    @abstractmethod
    def return_code_ok(self) -> int:
        pass

    @property
    @abstractmethod
    def return_code_wa(self) -> int:
        pass

    def __init__(
        self,
        env: Env,
        judge: RunSection,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        seed: int | None,
        expected_verdict: Verdict | None,
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            checker_name=judge.name,
            test=test,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_verdict=expected_verdict,
            **kwargs,
        )
        self.judge = judge
        self.seed = seed

    def _load_stderr(self) -> tuple[str, dict[str, Any]]:
        KEYS = {
            "POINTS": Decimal,
            "LOG": str,
            "NOTE": str,
        }
        FAIL_KWARGS = {
            "status": False,
            "stdout": False,
            "stderr_force_content": True,
        }

        key_values: dict[str, Any] = {}
        with self._open_file(self.checker_log_file) as f:
            message = f.readline().removesuffix("\n")
            if len(message.encode()) > 255:
                raise self._create_program_failure(
                    f"Message longer than 255 bytes: '{message}'",
                    self._result,
                    **FAIL_KWARGS,
                )
            for line in f.readlines():
                line = line.removesuffix("\n")

                if "=" not in line:
                    raise self._create_program_failure(
                        f"Invalid key-value pair: '{line}'", self._result, **FAIL_KWARGS
                    )
                key, val = line.split("=", 1)

                if key not in KEYS:
                    raise self._create_program_failure(
                        f"Invalid key: '{key}'", self._result, **FAIL_KWARGS
                    )
                if key in key_values:
                    raise self._create_program_failure(
                        f"Duplicate key: '{key}'", self._result, **FAIL_KWARGS
                    )
                if len(val.encode()) > 255:
                    raise self._create_program_failure(
                        f"Value longer than 255 bytes: '{val}'",
                        self._result,
                        **FAIL_KWARGS,
                    )
                try:
                    key_values[key] = KEYS[key](val)
                except ValueError:
                    raise self._create_program_failure(
                        f"Value is not of type '{KEYS[key]}': '{val}'",
                        self._result,
                        **FAIL_KWARGS,
                    )

        return (message, key_values)

    def _check(self) -> SolutionResult:
        envs = {}
        if self._env.config.tests.judge_needs_in:
            envs["TEST_INPUT"] = self.input.abspath
            self._access_file(self.input)
        if self._env.config.tests.judge_needs_out:
            envs["TEST_OUTPUT"] = self.correct_output.abspath
            self._access_file(self.correct_output)

        self._result = self._run_program(
            ProgramRole.judge,
            self.judge,
            args=[
                str(self.test),
                f"{self.seed:016x}" if self.seed is not None else OPENDATA_NO_SEED,
            ],
            stdin=self.output,
            stderr=self.checker_log_file,
            env=envs,
        )

        if self._result.returncode not in (self.return_code_ok, self.return_code_wa):
            # We need to check this first to report the correct problem
            raise self._create_program_failure(
                f"Judge failed on output {self.output:n}:",
                self._result,
                stderr_force_content=True,
            )

        message, key_values = self._load_stderr()

        if "LOG" in key_values:
            self._log(
                "info", f"Judge on output {self.output:p}: LOG={key_values['LOG']}"
            )

        if self._result.returncode == self.return_code_ok:
            if "POINTS" not in key_values:
                return RelativeSolutionResult(
                    verdict=Verdict.ok,
                    message=message,
                    solution_rr=self._solution_run_res,
                    checker_rr=self._result,
                    log=key_values.get("LOG"),
                    note=key_values.get("NOTE"),
                    relative_points=Decimal(1),
                )

            max_points = self._env.config.test_sections[self.test].points
            points = key_values["POINTS"]
            if max_points == "unscored" or points == max_points:
                verdict = Verdict.ok
            elif points < max_points:
                verdict = Verdict.partial_ok
            else:
                verdict = Verdict.superopt

            return AbsoluteSolutionResult(
                verdict=verdict,
                message=message,
                solution_rr=self._solution_run_res,
                checker_rr=self._result,
                log=key_values.get("LOG"),
                note=key_values.get("NOTE"),
                absolute_points=points,
            )

        else:
            return RelativeSolutionResult(
                verdict=Verdict.wrong_answer,
                message=message,
                solution_rr=self._solution_run_res,
                checker_rr=self._result,
                log=key_values.get("LOG"),
                note=key_values.get("NOTE"),
                relative_points=Decimal(0),
            )


class RunOpendataV1Judge(RunOpendataJudge):
    """Checks solution output using judge with the opendataV1 interface."""

    @property
    def return_code_ok(self) -> int:
        return 0

    @property
    def return_code_wa(self) -> int:
        return 1


class RunOpendataV2Judge(RunOpendataJudge):
    """Checks solution output using judge with the opendataV2 interface."""

    @property
    def return_code_ok(self) -> int:
        return 42

    @property
    def return_code_wa(self) -> int:
        return 43



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Iterable

from pisek.utils.paths import TaskPath, InputPath, OutputPath
from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.config.config_types import TaskType
from pisek.task_jobs.task_manager import TaskJobManager, GENERATOR_MAN_CODE
from pisek.task_jobs.data.testcase_info import (
    TestcaseInfo,
    TestcaseGenerationMode,
    ExportInputsList,
)

from .data import LinkData

TEST_SEED = 25265
SHORTEN_INPUTS_CUTOFF = 3


class DataManager(TaskJobManager):
    """Moves data to correct folders."""

    def __init__(self) -> None:
        super().__init__("Processing data")

    def _get_jobs(self) -> list[Job]:
        static_inputs = self._globs_to_files(
            ["*.in"], TaskPath.static_path(self._env, ".")
        )

        static_testcase_infos: list[TestcaseInfo] = []
        for input_path in static_inputs:
            name = input_path.name.removesuffix(".in")
            output_path = input_path.replace_suffix(".out")

            if (
                self._env.config.task.task_type == TaskType.interactive
                or output_path.exists()
            ):
                static_testcase_infos.append(TestcaseInfo.static(name))
            else:
                static_testcase_infos.append(TestcaseInfo.mixed(name))

        all_testcase_infos = list(static_testcase_infos)
        if self._env.config.tests.in_gen is not None:
            all_testcase_infos += self.prerequisites_results[GENERATOR_MAN_CODE][
                "inputs"
            ]

        all_testcase_infos.sort(key=lambda info: info.name)

        # put inputs in tests
        self._testcase_infos: dict[int, list[TestcaseInfo]] = {}

        for test in self._env.config.test_sections.values():
            self._testcase_infos[test.num] = []

            for testcase_info in all_testcase_infos:
                inp_path = testcase_info.input_path(TEST_SEED).name
                if test.in_test(inp_path):
                    self._testcase_infos[test.num].append(testcase_info)

            if len(self._testcase_infos[test.num]) == 0:
                raise PipelineItemFailure(
                    f"No inputs for test {test.num} with globs {test.all_globs}."
                )

        for testcase_info in self._testcase_infos[0]:
            if testcase_info.generation_mode != TestcaseGenerationMode.static:
                raise PipelineItemFailure(
                    f"Sample inputs must be static, but '{testcase_info.name}' is {testcase_info.generation_mode}."
                )

        used_inputs = set(sum(self._testcase_infos.values(), start=[]))
        self._report_not_included_inputs(
            used_inputs - set(self._testcase_infos[self._env.config.tests_count - 1])
        )
        self._report_unused_inputs(set(all_testcase_infos) - used_inputs)
        self._check_one_input_in_nonsample_test()

        jobs: list[Job] = []

        for testcase in sorted(used_inputs, key=lambda i: i.name):
            name = testcase.name
            mode = testcase.generation_mode

            if mode in (TestcaseGenerationMode.static, TestcaseGenerationMode.mixed):
                input_target_path = InputPath.new(f"{name}.in")
                jobs.append(
                    LinkData(
                        self._env,
                        TaskPath.static_path(self._env, f"{name}.in"),
                        input_target_path.to_raw(self._env.config.tests.in_format),
                    )
                )
            if (
                mode == TestcaseGenerationMode.static
                and self._env.config.task.task_type != TaskType.interactive
            ):
                output_target_path = OutputPath.static(f"{name}.out")
                jobs.append(
                    LinkData(
                        self._env,
                        TaskPath.static_path(self._env, f"{name}.out"),
                        output_target_path.to_raw(self._env.config.tests.out_format),
                    )
                )

        jobs.append(ExportInputsList(self._env, list(sorted(used_inputs))))

        if self._env.config.tests.validator is None:
            not_skipped_test = any(
                test_config.checks_validate
                for test_config in self._env.config.test_sections.values()
            )
            if not_skipped_test:
                self._warn("No validator specified in config.")

        return jobs

    def _report_unused_inputs(self, unused_inputs: Iterable[TestcaseInfo]) -> None:
        inputs = list(sorted(unused_inputs, key=lambda inp: inp.name))
        if self._env.config.checks.no_unused_inputs and inputs:
            if self._env.verbosity <= 0:
                self._warn(
                    f"{len(inputs)} unused input{'s' if len(inputs) >= 2 else ''}. "
                    f"({self._short_inputs_list(inputs)})"
                )
            else:
                for inp in inputs:
                    self._warn(
                        f"Unused {'generated' if inp.generation_mode == TestcaseGenerationMode.generated else 'static'} input: '{inp.name}.in'"
                    )

    def _report_not_included_inputs(
        self, not_included_inputs: Iterable[TestcaseInfo]
    ) -> None:
        inputs = list(sorted(not_included_inputs, key=lambda inp: inp.name))
        if self._env.config.checks.all_inputs_in_last_test and inputs:
            if self._env.verbosity <= 0:
                self._warn(
                    f"{len(inputs)} input{'s' if len(inputs) >= 2 else ''} "
                    "not included in last test. "
                    f"({self._short_inputs_list(inputs)})"
                )
            else:
                for inp in inputs:
                    self._warn(f"Input '{inp.name}.in' not included in last test.")

    def _check_one_input_in_nonsample_test(self) -> None:
        if not self._env.config.checks.one_input_in_each_nonsample_test:
            return

        for test_num, testcases in self._testcase_infos.items():
            if test_num == 0:
                continue
            cnt = sum(tc.repeat for tc in testcases)
            if cnt != 1:
                self._warn(
                    f"{self._env.config.test_sections[test_num].name} contains {cnt} testcases but should contain 1."
                )

    def _short_inputs_list(self, inputs: Iterable[TestcaseInfo]) -> str:
        return self._short_list(
            list(map(lambda inp: f"{inp.name}.in", inputs)), SHORTEN_INPUTS_CUTOFF
        )

    def _compute_result(self) -> dict[str, Any]:
        res = {"testcase_info": self._testcase_infos}
        return res



#!/usr/bin/env python3
import itertools
import random
import sys

FEATURES = ["few", "small", "equal", "max"]


def gen(input_name: str) -> None:
    max_n = 10 if "few" in input_name else 10**5
    max_k = 10**5 if "small" in input_name else 10**9

    if "max" in input_name:
        n = max_n
        k = max_k
    else:
        n = random.randint(1, max_n)
        k = random.randint(n, max_k)

    if "equal" in input_name:
        k = n

    print(n, k)


# No arguments - List all inputs we can generate
if len(sys.argv) == 1:
    for variant in itertools.product([True, False], repeat=4):
        features = [f for i, f in enumerate(FEATURES) if variant[i]]
        if "equal" in features and "small" not in features:
            continue  # Any equal input is small

        if features:
            print("_".join(features), end=" ")
            if "max" in features:
                # max is deterministic - no need for seed
                print("seeded=false")
            else:
                # Otherwise generate this input three times
                print("repeat=3")
        else:
            print("big repeat=5")

# Generate an input
else:
    # Input is seeded
    if len(sys.argv) == 3:
        random.seed(sys.argv[2])

    gen(sys.argv[1])



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
import logging
import os
import tempfile
from typing import Optional, Any, Union, Callable
import signal
import subprocess

from pisek.config.task_config import ProgramRole, RunSection
from pisek.env.env import Env
from pisek.utils.paths import TaskPath, LogPath
from pisek.jobs.jobs import PipelineItemFailure
from pisek.utils.text import tab
from pisek.task_jobs.run_result import RunResultKind, RunResult
from pisek.task_jobs.task_job import TaskJob

logger = logging.getLogger(__name__)


@dataclass
class ProgramPoolItem:
    executable: TaskPath
    args: list[str]
    time_limit: float
    clock_limit: float
    mem_limit: int
    process_limit: int
    stdin: Optional[Union[TaskPath, int]]
    stdout: Optional[Union[TaskPath, int]]
    stderr: Optional[TaskPath]
    env: dict[str, str] = field(default_factory=lambda: {})

    def to_popen(self, minibox: str, meta_file: str) -> dict[str, Any]:
        """Returns subprocess.Popen args for executing this PoolItem."""
        result: dict[str, Any] = {}

        minibox_args = []
        minibox_args.append(f"--time={self.time_limit}")
        minibox_args.append(f"--wall-time={self.clock_limit}")
        minibox_args.append(f"--mem={self.mem_limit*1024}")
        minibox_args.append(f"--processes={self.process_limit}")

        for std in ("stdin", "stdout", "stderr"):
            attr = getattr(self, std)
            if isinstance(attr, TaskPath):
                minibox_args.append(f"--{std}={attr.abspath}")
            elif getattr(self, std) is None:
                minibox_args.append(f"--{std}=/dev/null")

            if isinstance(attr, int):
                result[std] = attr
            else:
                result[std] = subprocess.PIPE

        for key, val in self.env.items():
            minibox_args.append(f"--env={key}={val}")

        minibox_args.append("--silent")
        minibox_args.append(f"--meta={meta_file}")

        result["args"] = (
            [minibox]
            + minibox_args
            + ["--run", "--", self.executable.abspath]
            + self.args
        )
        return result


class ProgramsJob(TaskJob):
    """Job that deals with a program."""

    def __init__(self, env: Env, name: str, **kwargs) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self._program_pool: list[ProgramPoolItem] = []
        self._callback: Optional[Callable[[subprocess.Popen], None]] = None

    @staticmethod
    def _env_disjoint_union(
        env1: dict[str, str], env2: dict[str, str]
    ) -> dict[str, str]:
        union = env1 | env2
        if len(union) != len(env1) + len(env2):
            duplicates = list(set(env1.keys()) & set(env2.keys()))
            raise PipelineItemFailure(
                f"Environment variable '{duplicates[0]}' is used by the protocol and cannot be overwritten."
            )
        return union

    def _load_executable(
        self,
        executable: TaskPath,
        args: list[str],
        time_limit: float,
        clock_limit: float,
        mem_limit: int,
        process_limit: int,
        stdin: Optional[Union[TaskPath, int]] = None,
        stdout: Optional[Union[TaskPath, int]] = None,
        stderr: Optional[LogPath] = None,
        env: dict[str, str] = {},
    ):
        if self._is_file(executable):
            self._access_file(executable)
        elif self._is_dir(executable):
            self._access_dir(executable)
            executable = executable.join("run")

        if self._is_dir(executable):
            raise PipelineItemFailure(
                f"Cannot execute '{executable:p}': Is a directory.."
            )
        elif not self._exists(executable):
            raise PipelineItemFailure(f"Cannot execute '{executable:p}': File missing.")

        if isinstance(stdin, TaskPath):
            self._access_file(stdin)
        if isinstance(stdout, TaskPath):
            self.make_filedirs(stdout)
            self._access_file(stdout)
        if isinstance(stderr, TaskPath):
            self.make_filedirs(stderr)
            self._access_file(stderr)

        self._program_pool.append(
            ProgramPoolItem(
                executable=executable,
                args=args,
                time_limit=time_limit,
                clock_limit=clock_limit,
                mem_limit=mem_limit,
                process_limit=process_limit,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                env=env,
            )
        )

    def _load_program(
        self,
        program_role: ProgramRole,
        program: RunSection,
        args: list[str] = [],
        stdin: Optional[Union[TaskPath, int]] = None,
        stdout: Optional[Union[TaskPath, int]] = None,
        stderr: Optional[LogPath] = None,
        env={},
    ) -> None:
        """Adds program to execution pool."""
        time_limit: Optional[float] = None
        if program_role.is_solution():
            time_limit = self._env.time_limit

        self._load_executable(
            executable=program.executable,
            args=program.args + args,
            time_limit=program.time_limit if time_limit is None else time_limit,
            clock_limit=program.clock_limit(time_limit),
            mem_limit=program.mem_limit,
            process_limit=program.process_limit,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=self._env_disjoint_union(env, program.env),
        )

    def _load_callback(self, callback: Callable[[subprocess.Popen], None]) -> None:
        if self._callback is not None:
            raise RuntimeError("Callback already loaded.")
        self._callback = callback

    def _run_programs(self) -> list[RunResult]:
        """Runs all programs in execution pool."""
        running_pool: list[subprocess.Popen] = []
        meta_files: list[str] = []
        tmp_dirs: list[str] = []
        minibox = TaskPath.executable_path("_minibox").abspath
        for pool_item in self._program_pool:
            fd, meta_file = tempfile.mkstemp()
            os.close(fd)
            meta_files.append(meta_file)

            popen = pool_item.to_popen(minibox, meta_file)
            logger.debug("Executing '" + " ".join(popen["args"]) + "'")

            tmp_dir = tempfile.mkdtemp(prefix="pisek_")
            tmp_dirs.append(tmp_dir)
            running_pool.append(subprocess.Popen(**popen, cwd=tmp_dir))

        if self._callback is not None:
            callback_exec = False
            while True:
                states = [process.poll() is not None for process in running_pool]
                if not callback_exec and any(states):
                    callback_exec = True
                    self._callback(running_pool[states.index(True)])

                if all(states):
                    break

        run_results = []
        for pool_item, process, tmp_dir, meta_file in zip(
            self._program_pool, running_pool, tmp_dirs, meta_files
        ):
            self._wait_for_subprocess(process)
            assert process.stderr is not None  # To make mypy happy

            with open(meta_file) as f:
                meta_raw = f.read().strip().split("\n")

            assert meta_file.startswith("/tmp")  # Better safe then sorry
            assert tmp_dir.startswith("/tmp")
            os.remove(meta_file)
            os.removedirs(tmp_dir)

            meta = {key: val for key, val in map(lambda x: x.split(":", 1), meta_raw)}
            if process.returncode == 0:
                t, wt = float(meta["time"]), float(meta["time-wall"])
                run_results.append(
                    RunResult(
                        RunResultKind.OK,
                        0,
                        t,
                        wt,
                        pool_item.stdin,
                        pool_item.stdout,
                        pool_item.stderr,
                        "Exited with return code 0",
                    )
                )
            elif process.returncode == 1:
                t, wt = float(meta["time"]), float(meta["time-wall"])
                if meta["status"] in ("RE", "SG"):
                    if meta["status"] == "RE":
                        return_code = int(meta["exitcode"])
                    elif meta["status"] == "SG":
                        return_code = int(meta["exitsig"])
                        meta["message"] += f" ({signal.Signals(return_code).name})"

                    run_results.append(
                        RunResult(
                            RunResultKind.RUNTIME_ERROR,
                            return_code,
                            t,
                            wt,
                            pool_item.stdin,
                            pool_item.stdout,
                            pool_item.stderr,
                            meta["message"],
                        )
                    )
                elif meta["status"] == "TO":
                    time_limit = (
                        f"{pool_item.time_limit}s"
                        if t > pool_item.time_limit
                        else f"{pool_item.clock_limit}ws"
                    )
                    run_results.append(
                        RunResult(
                            RunResultKind.TIMEOUT,
                            -1,
                            t,
                            wt,
                            pool_item.stdin,
                            pool_item.stdout,
                            pool_item.stderr,
                            f"Timeout after {time_limit}",
                        )
                    )
                else:
                    raise RuntimeError(f"Unknown minibox status {meta['message']}.")
            else:
                raise PipelineItemFailure(
                    f"Minibox error:\n{tab(process.stderr.read().decode())}"
                )

        return run_results

    def _run_program(
        self,
        program_role: ProgramRole,
        program: RunSection,
        **kwargs,
    ) -> RunResult:
        """Loads one program and runs it."""
        self._load_program(program_role, program, **kwargs)
        return self._run_programs()[0]

    def _run_tool(
        self,
        program: str,
        args: list[str] = [],
        stdin: Optional[Union[TaskPath, int]] = None,
        stdout: Optional[Union[TaskPath, int]] = None,
        stderr: Optional[LogPath] = None,
        env={},
    ) -> RunResult:
        """Loads one program and runs it."""
        self._load_executable(
            TaskPath.executable_path(program),
            args=args,
            time_limit=300,
            clock_limit=300,
            mem_limit=0,
            process_limit=1,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=env,
        )
        return self._run_programs()[0]

    def _create_program_failure(self, msg: str, res: RunResult, **kwargs):
        """Create PipelineItemFailure that nicely formats RunResult"""
        return PipelineItemFailure(
            f"{msg}\n{tab(self._format_run_result(res, **kwargs))}"
        )

    def _format_run_result(
        self,
        res: RunResult,
        status: bool = True,
        stdin: bool = True,
        stdin_force_content: bool = False,
        stdout: bool = True,
        stdout_force_content: bool = False,
        stderr: bool = True,
        stderr_force_content: bool = False,
        time: bool = False,
    ):
        """Formats RunResult."""
        program_msg = ""
        if status:
            program_msg += f"status: {res.status}\n"

        if stdin and isinstance(res.stdin_file, TaskPath):
            program_msg += f"stdin: {self._quote_file_with_name(res.stdin_file, force_content=stdin_force_content)}"
        if stdout and isinstance(res.stdout_file, TaskPath):
            program_msg += f"stdout: {self._quote_file_with_name(res.stdout_file, force_content=stdout_force_content)}"
        if stderr and isinstance(res.stderr_file, TaskPath):
            program_msg += f"stderr: {self._quote_file_with_name(res.stderr_file, force_content=stderr_force_content, style='ht')}"
        if time:
            program_msg += f"time: {res.time}\n"

        return program_msg.removesuffix("\n")



#!/usr/bin/env python3
# Does not generate any tests for the third test.

import sys
import os

test_dir = sys.argv[1]
os.makedirs(test_dir, exist_ok=True)

with open(os.path.join(test_dir, "01_01.in"), "w") as f:
    f.write("1 2")

with open(os.path.join(test_dir, "03_01.in"), "w") as f:
    f.write("-5 2")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from enum import auto, Enum
from functools import partial, cache, cached_property
from typing import Callable, Optional, TYPE_CHECKING

from pisek.utils.colors import ColorSettings
from pisek.config.config_types import TestPoints
from pisek.task_jobs.run_result import RunResult

if TYPE_CHECKING:
    from pisek.env.env import Env


class Verdict(Enum):
    # Higher value means more unsuccessful verdict.
    superopt = auto()
    ok = auto()
    partial_ok = auto()
    timeout = auto()
    wrong_answer = auto()
    normalization_fail = auto()
    error = auto()

    def is_zero_point(self) -> bool:
        return self in (
            Verdict.timeout,
            Verdict.wrong_answer,
            Verdict.error,
            Verdict.normalization_fail,
        )

    @cached_property
    def mark(self) -> str:
        mark = {
            Verdict.ok: "·",
            Verdict.superopt: "S",
            Verdict.partial_ok: "P",
            Verdict.timeout: "T",
            Verdict.wrong_answer: "W",
            Verdict.error: "!",
            Verdict.normalization_fail: "N",
        }[self]
        return ColorSettings.colored(mark, self.color)

    @cached_property
    def color(self) -> str:
        return {
            Verdict.ok: "green",
            Verdict.superopt: "blue",
            Verdict.partial_ok: "yellow",
            Verdict.timeout: "red",
            Verdict.wrong_answer: "red",
            Verdict.error: "red",
            Verdict.normalization_fail: "red",
        }[self]

    @staticmethod
    @cache
    def pad_length() -> int:
        return max(len(v.name) for v in Verdict)


@dataclass(init=False)
class SolutionResult(ABC):
    """Class representing result of a solution on given input."""

    verdict: Verdict
    message: str | None
    solution_rr: RunResult
    checker_rr: RunResult | None
    log: str | None = None
    note: str | None = None

    def points(self, env: "Env", test_points: TestPoints) -> Decimal:
        if test_points == "unscored":
            return Decimal("0")
        else:
            return self._points(env, test_points)

    @abstractmethod
    def _points(self, env: "Env", test_points: int) -> Decimal:
        pass


@dataclass(kw_only=True)
class RelativeSolutionResult(SolutionResult):
    verdict: Verdict
    message: Optional[str]
    solution_rr: RunResult
    checker_rr: Optional[RunResult]
    log: str | None = None
    note: str | None = None
    relative_points: Decimal

    def _points(self, env: "Env", test_points: int) -> Decimal:
        return (self.relative_points * test_points).quantize(
            Decimal("0.1") ** env.config.task.score_precision
        )


@dataclass(kw_only=True)
class AbsoluteSolutionResult(SolutionResult):
    verdict: Verdict
    message: Optional[str]
    solution_rr: RunResult
    checker_rr: Optional[RunResult]
    log: str | None = None
    note: str | None = None
    absolute_points: Decimal

    def _points(self, env: "Env", test_points: int) -> Decimal:
        return self.absolute_points


def verdict_always(res: Verdict) -> bool:
    return True


def verdict_accepted(res: Verdict) -> bool:
    return res in (Verdict.ok, Verdict.superopt)


def specific_verdict(res: Verdict, verdict: Verdict) -> bool:
    return res == verdict


# Specifies how expected str should be interpreted
# First function must be true for all
# Second function must be true for any
TEST_SPEC: dict[str, tuple[Callable[[Verdict], bool], Callable[[Verdict], bool]]] = {
    "1": (
        verdict_accepted,
        partial(specific_verdict, verdict=Verdict.ok),
    ),
    "S": (partial(specific_verdict, verdict=Verdict.superopt), verdict_always),
    "A": (verdict_accepted, verdict_always),
    "0": (verdict_always, lambda verdict: verdict.is_zero_point()),
    "X": (verdict_always, verdict_always),
    "P": (
        lambda verdict: not verdict.is_zero_point(),
        partial(specific_verdict, verdict=Verdict.partial_ok),
    ),
    "W": (
        verdict_always,
        partial(specific_verdict, verdict=Verdict.wrong_answer),
    ),
    "N": (
        verdict_always,
        partial(specific_verdict, verdict=Verdict.normalization_fail),
    ),
    "!": (verdict_always, partial(specific_verdict, verdict=Verdict.error)),
    "T": (verdict_always, partial(specific_verdict, verdict=Verdict.timeout)),
}



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pisek.jobs.job_pipeline import JobPipeline
from pisek.env.env import Env, TestingTarget
from pisek.config.config_types import TaskType, OutCheck
from pisek.utils.paths import InputPath
from pisek.task_jobs.task_manager import (
    TOOLS_MAN_CODE,
    INPUTS_MAN_CODE,
    BUILD_MAN_CODE,
    GENERATOR_MAN_CODE,
    FUZZ_MAN_CODE,
    SOLUTION_MAN_CODE,
)

from pisek.jobs.jobs import JobManager
from pisek.task_jobs.tools import ToolsManager
from pisek.task_jobs.data.data_manager import DataManager
from pisek.task_jobs.generator.generator_manager import (
    PrepareGenerator,
    RunGenerator,
    TestcaseInfoMixin,
)
from pisek.task_jobs.checker.fuzzing_manager import FuzzingManager
from pisek.task_jobs.builder.build import BuildManager
from pisek.task_jobs.solution.ui_managers import EmptyLineManager, TestsHeaderManager
from pisek.task_jobs.solution.solution_manager import SolutionManager
from pisek.task_jobs.testing_log import CreateTestingLog
from pisek.task_jobs.completeness_check import CompletenessCheck


class TaskPipeline(JobPipeline):
    """JobPipeline that checks whether task behaves as expected."""

    def __init__(self, env: Env):
        super().__init__()
        self.job_managers = [job_man for job_man, _ in self._get_named_pipeline(env)]

    def _get_named_pipeline(self, env: Env) -> list[tuple[JobManager, str]]:
        named_pipeline: list[tuple[JobManager, str]] = [
            tools := (ToolsManager(), TOOLS_MAN_CODE),
            build := (BuildManager(), BUILD_MAN_CODE),
        ]
        build[0].add_prerequisite(*tools)
        if env.config.tests.in_gen is not None:
            named_pipeline.append(generator := (PrepareGenerator(), GENERATOR_MAN_CODE))
            generator[0].add_prerequisite(*build)
        named_pipeline.append(inputs := (DataManager(), INPUTS_MAN_CODE))

        inputs[0].add_prerequisite(*build)
        if env.config.tests.in_gen is not None:
            inputs[0].add_prerequisite(*generator)

        if env.target == TestingTarget.build:
            return named_pipeline

        solutions = []
        self.input_generator: TestcaseInfoMixin

        if env.target == TestingTarget.generator or not env.config.solutions:
            named_pipeline.append(gen_inputs := (RunGenerator(), ""))
            gen_inputs[0].add_prerequisite(*inputs)
            self.input_generator = gen_inputs[0]

        else:
            # First solution generates inputs
            assert (
                not env.config.tests.judge_needs_out
                or env.solutions[0] == env.config.primary_solution
            )

            named_pipeline.append((EmptyLineManager(), ""))
            if env.verbosity == 0:
                named_pipeline.append(th := (TestsHeaderManager(), ""))
                th[0].add_prerequisite(*inputs)

            named_pipeline.append(
                first_solution := (
                    SolutionManager(env.solutions[0], True),
                    f"{SOLUTION_MAN_CODE}{env.solutions[0]}",
                )
            )
            solutions.append(first_solution)
            self.input_generator = first_solution[0]

            for sol_name in env.solutions[1:]:
                named_pipeline.append(
                    solution := (
                        SolutionManager(sol_name, False),
                        f"{SOLUTION_MAN_CODE}{sol_name}",
                    )
                )

                solution[0].add_prerequisite(*first_solution)
                solutions.append(solution)

            for solution in solutions:
                solution[0].add_prerequisite(*inputs)

            if env.verbosity == 0:
                named_pipeline.append((EmptyLineManager(), ""))

        if env.testing_log:
            named_pipeline.append(testing_log := (CreateTestingLog(), ""))
            for solution in solutions:
                testing_log[0].add_prerequisite(*solution)

        fuzz_judge: tuple[Optional[FuzzingManager], Optional[str]]
        if (
            env.target == TestingTarget.all
            and env.config.solutions
            and env.config.task.task_type != TaskType.interactive
            and env.config.tests.out_check == OutCheck.judge
            and (
                env.config.checks.fuzzing_thoroughness > 0
                or env.config.checks.judge_rejects_trailing_string
            )
        ):
            named_pipeline.append(fuzz_judge := (FuzzingManager(), FUZZ_MAN_CODE))
            fuzz_judge[0].add_prerequisite(*first_solution)
        else:
            fuzz_judge = (None, None)

        if solutions and env.target == TestingTarget.all:
            named_pipeline.append(completeness_check := (CompletenessCheck(), ""))
            completeness_check[0].add_prerequisite(*fuzz_judge)
            for solution in solutions:
                completeness_check[0].add_prerequisite(*solution)

        return named_pipeline

    def input_dataset(self) -> list[InputPath]:
        if self.input_generator.result is None:
            raise RuntimeError("Input dataset has not been computed yet.")
        return self.input_generator.result["input_dataset"]



#!/usr/bin/env python3
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Nacte ze stdin vstupy ulohy a zkontroluje jejich spravnost"
    )
    parser.add_argument("test", type=int, help="test (indexovany od 1)")
    args = parser.parse_args()

    # Here the bounds for test 3 are stricter than what the generator really creates.
    BOUNDS = [(-1e18, 1e18), (0, 1e9), (-1e9, 1e9), (-1e9, 1e9)][args.test]
    a, b = map(int, input().split(" "))
    assert BOUNDS[0] <= a <= BOUNDS[1]
    assert BOUNDS[0] <= b <= BOUNDS[1]

    end = True
    try:
        input()
    except EOFError:
        end = True
    assert end
    exit(42)



#!/usr/bin/env python3
# Generates inputs that all have the same answer

import random
import sys
import os

test_dir = sys.argv[1]
os.makedirs(test_dir, exist_ok=True)

ANS = 12345

POSITIVE_ONLY = [True, False, False]
MAX_ABS = [int(1e9), int(1e9), int(1e18)]

random.seed(123)

for test_i, (positive_only, max_abs) in enumerate(zip(POSITIVE_ONLY, MAX_ABS)):
    for ti in range(5):

        if positive_only:
            a = random.randint(0, ANS)
        else:
            a = random.randint(
                max(-max_abs, ANS - max_abs), min(max_abs, ANS + max_abs)
            )

        test_filename = "{:0>2}_{:0>2}.in".format(test_i + 1, ti + 1)
        with open(os.path.join(test_dir, test_filename), "w") as f:
            f.write(f"{a} {ANS-a}\n")



# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cms.db.task import Dataset, Testcase
from cms.db.filecacher import FileCacher
from sqlalchemy.orm import Session
from glob import glob
from itertools import chain

from pisek.env.env import Env
from pisek.config.config_types import TaskType
from pisek.utils.paths import TaskPath


def create_testcase(
    session: Session,
    files: FileCacher,
    dataset: Dataset,
    codename: str,
    input_file: TaskPath,
    output_file: TaskPath | None,
) -> Testcase:
    input = files.put_file_from_path(input_file.path, f"Input for testcase {codename}")

    if output_file is not None:
        output = files.put_file_from_path(
            output_file.path, f"Output for testcase {codename}"
        )
    else:
        output = files.put_file_content(
            "No output".encode(), "Almost empty output file"
        )

    testcase = Testcase(
        dataset=dataset, codename=codename, input=input, output=output, public=True
    )

    session.add(testcase)
    return testcase
